from collections import Callable

from inoft_vocal_framework.platforms_handlers.alexa_v1.audioplayer.audioplayer_directives import AudioPlayerWrapper
from inoft_vocal_framework.platforms_handlers.alexa_v1.context import Context
from inoft_vocal_framework.platforms_handlers.alexa_v1.request import Request
from inoft_vocal_framework.platforms_handlers.alexa_v1.response.response import Response
from inoft_vocal_framework.platforms_handlers.alexa_v1.session import Session
from inoft_vocal_framework.safe_dict import SafeDict


class AlexaHandlerInput:
    from inoft_vocal_framework.platforms_handlers.handler_input import HandlerInput

    def __init__(self, parent_handler_input: HandlerInput):
        self.parent_handler_input = parent_handler_input
        self.session = Session()
        self.context = Context()
        self.request = Request()
        self.response = Response()

        self._is_new_session = None
        self._session_attributes = None
        self._audioplayer = None

    @property
    def is_new_session(self) -> bool:
        if self._is_new_session is None:
            self._is_new_session = self.session.new if isinstance(self.session.new, bool) else False
        return self._is_new_session

    def is_launch_request(self) -> bool:
        return self.request.is_launch_request()

    def is_in_intent_names(self, intent_names_list) -> bool:
        return self.request.is_in_intent_names(intent_names_list=intent_names_list)

    def is_in_request_types(self, request_types_list: list):
        return self.request.is_in_request_types(request_types_list=request_types_list)

    # todo: create a new function that handle the end of a session (like an optionnal function in each class type ?)
    def say(self, text_or_ssml: str) -> None:
        self.response.say(text_or_ssml=text_or_ssml)

    def reprompt(self, text_or_ssml: str) -> None:
        # todo: fix reprompt (for alexa and dialogflow)
        self.response.say_reprompt(text_or_ssml=text_or_ssml)

    @property
    def audioplayer(self) -> AudioPlayerWrapper:
        if self._audioplayer is None:
            self._audioplayer = AudioPlayerWrapper(parent_handler_input=self.parent_handler_input)
        return self._audioplayer

    def save_audioplayer_handlers_group_class(self, handlers_group_class_type: type, group_class_kwargs: dict = None):
        from inspect import getfile
        # We use the persistent attributes and not the session, because after launching an audio file with the audio player,
        # the session of the user will end. Then when interacting with an audio file there will be no session id.
        # So, if we save this data as session attributes, it would be considered of the same session only if the smart
        # session timeout has not been exceeded. Which is not at all what we want.
        self.parent_handler_input.persistent_memorize(data_key="lastUsedAudioPlayerHandlersGroupClass",
                                                      data_value={
                                                          "fileFilepathContainingClass": getfile(handlers_group_class_type),
                                                          "classPath": handlers_group_class_type.__qualname__,
                                                          "classKwargs": group_class_kwargs,
                                                      })

    def get_last_used_audioplayer_handlers_group(self) -> SafeDict:
        return SafeDict(self.parent_handler_input.persistent_remember("lastUsedAudioPlayerHandlersGroupClass", specific_object_type=dict))

    def show_basic_card(self, title: str, text: str, small_image_url: str = None, large_image_url: str = None) -> None:
        from inoft_vocal_framework.platforms_handlers.alexa_v1.response.response import Card
        if small_image_url is not None or large_image_url is not None:
            from inoft_vocal_framework.platforms_handlers.alexa_v1.response.response import Image

            if small_image_url is None and large_image_url is not None:
                small_image_url = large_image_url
                print("WARNING ! The small_image_url argument was not specified, the large_image_url"
                      "argument (which has been specified) is now also used as the small_image_url.")
            elif large_image_url is None and small_image_url is not None:
                large_image_url = small_image_url
                print("WARNING ! The large_image_url argument was not specified, the small_image_url"
                      "argument (which has been specified) is now also used as the large_image_url.")

            self.response.card = Card(type_value=Card.type_standard, title=title, text=text,
                                      image=Image(small_image_url=small_image_url, large_image_url=large_image_url))
        else:
            self.response.card = Card(type_value=Card.type_simple, title=title, content_text=text)

    def show_link_account_card(self) -> None:
        raise Exception("Not yet implemented")

    def show_ask_permissions_card(self) -> None:
        raise Exception("Not yet implemented")

    def end_session(self, should_end: bool = True) -> None:
        self.response.end_session(should_end=should_end)

    @property
    def session_attributes(self) -> SafeDict:
        if not isinstance(self._session_attributes, SafeDict):
            if isinstance(self.session.attributes, dict):
                self._session_attributes = SafeDict(self.session.attributes)
            else:
                self._session_attributes = SafeDict()
        return self._session_attributes




from inoft_vocal_framework.exceptions import raise_if_variable_not_expected_type
from inoft_vocal_framework.platforms_handlers.nested_object_to_dict import NestedObjectToDict
from inoft_vocal_framework.safe_dict import SafeDict


class Intent:
    json_key = "intent"

    def __init__(self):
        self._name = str()
        self._displayName = str()

    @property
    def name(self) -> str:
        return self._name

    @name.setter
    def name(self, name: str) -> None:
        if not isinstance(name, str):
            raise Exception(f"name was type {type(name)} which is not valid value for his parameter.")
        self._name = name

    @property
    def displayName(self) -> str:
        return self._displayName

    @displayName.setter
    def displayName(self, displayName: str) -> None:
        if not isinstance(displayName, str):
            raise Exception(f"displayName was type {type(displayName)} which is not valid value for his parameter.")
        self._displayName = displayName

class User:
    json_key = "user"

    VERIFICATION_NAME_GUEST = "GUEST"
    VERIFICATION_NAME_VERIFIED = "VERIFIED"

    def __init__(self):
        self._locale = str()
        self._lastSeen = str()
        self._userStorage = str()
        self._userVerificationStatus = str()

    @property
    def locale(self) -> str:
        return self._locale

    @locale.setter
    def locale(self, locale: str) -> None:
        if not isinstance(locale, str):
            raise Exception(f"locale was type {type(locale)} which is not valid value for his parameter.")
        self._locale = locale

    @property
    def lastSeen(self) -> str:
        return self._lastSeen

    @lastSeen.setter
    def lastSeen(self, lastSeen: str) -> None:
        if not isinstance(lastSeen, str):
            raise Exception(f"lastSeen was type {type(lastSeen)} which is not valid value for his parameter.")
        self._lastSeen = lastSeen

    @property
    def userStorage(self) -> str:
        return self._userStorage

    @userStorage.setter
    def userStorage(self, userStorage: str) -> None:
        if not isinstance(userStorage, str):
            raise Exception(f"userStorage was type {type(userStorage)} which is not valid value for his parameter.")
        self._userStorage = userStorage

    @property
    def userVerificationStatus(self) -> str:
        return self._userVerificationStatus

    @userVerificationStatus.setter
    def userVerificationStatus(self, userVerificationStatus: str) -> None:
        if not isinstance(userVerificationStatus, str):
            raise Exception(f"userVerificationStatus was type {type(userVerificationStatus)} which is not valid value for his parameter.")
        self._userVerificationStatus = userVerificationStatus


class Payload:
    json_key = "payload"
    INPUT_TYPE_OPTION = "OPTION"

    def __init__(self):
        self._user = User()
        self._conversation = self.Conversation()
        self._inputs = self.InputsCustomList()
        self._surface = self.Surface()
        self._availableSurfaces = self.AvailableSurfaces()
        self._isInSandbox = bool()
        self._requestType = str()

    def get_first_input_of_type(self, type_name: str):
        for input_item in self.inputs:
            for argument_item in input_item.arguments:
                if argument_item.name == type_name:
                    return argument_item

    @property
    def user(self) -> User:
        return self._user

    class Conversation:
        json_key = "conversation"

        def __init__(self):
            self._conversationId = str()
            self._type = str()

        @property
        def conversationId(self) -> str:
            return self._conversationId

        @conversationId.setter
        def conversationId(self, conversationId: str) -> None:
            raise_if_variable_not_expected_type(value=conversationId, expected_type=str, variable_name="conversationId")
            self._conversationId = conversationId

        @property
        def type(self) -> str:
            return self._type

        @type.setter
        def type(self, type_key: str) -> None:
            raise_if_variable_not_expected_type(value=type_key, expected_type=str, variable_name="type_key")
            self._type = type

    @property
    def conversation(self) -> Conversation:
        return self._conversation

    class InputsCustomList(list):
        # todo: make the check that the current device has the capabilities to use an interactive list
        def __init__(self):
            super().__init__()

        def append(self, item) -> None:
            if isinstance(item, dict):
                input_item_object = self.InputItem()
                NestedObjectToDict.process_and_set_json_to_object(object_class_to_set_to=input_item_object,
                    request_json_dict_stringed_dict_or_list=item, key_names_identifier_objects_to_go_into=["json_key"])
                super().append(input_item_object)

        def custom_set_from(self, list_object: list) -> None:
            for item in list_object:
                self.append(item=item)

        class InputItem:
            def __init__(self):
                self._intent = str()
                self._rawInputs = list()
                self._arguments = self.ArgumentItemsCustomList()

            @property
            def intent(self) -> str:
                return self._intent

            def intent(self, intent: str) -> None:
                raise_if_variable_not_expected_type(value=intent, expected_type=str, variable_name="intent")
                self._intent = intent

            @property
            def rawInputs(self) -> list:
                return self._rawInputs

            @rawInputs.setter
            def rawInputs(self, rawInputs: list) -> None:
                raise_if_variable_not_expected_type(value=rawInputs, expected_type=list, variable_name="rawInputs")
                self._rawInputs = rawInputs

            class ArgumentItemsCustomList(list):
                def __init__(self):
                    super().__init__()

                def append(self, item) -> None:
                    if isinstance(item, dict):
                        argument_item_object = self.ArgumentItem()
                        NestedObjectToDict.process_and_set_json_to_object(object_class_to_set_to=argument_item_object,
                            request_json_dict_stringed_dict_or_list=item, key_names_identifier_objects_to_go_into=["json_key"])
                        super().append(argument_item_object)

                def custom_set_from(self, list_object: list) -> None:
                    for item in list_object:
                        self.append(item=item)

                class ArgumentItem:
                    json_key = None

                    def __init__(self):
                        self._name = str()
                        self._textValue = str()
                        self._rawText = str()

                    @property
                    def name(self) -> str:
                        return self._name

                    @name.setter
                    def name(self, name: str) -> None:
                        raise_if_variable_not_expected_type(value=name, expected_type=str, variable_name="name")
                        self._name = name

                    @property
                    def textValue(self) -> str:
                        return self._textValue

                    @textValue.setter
                    def textValue(self, textValue: str) -> None:
                        raise_if_variable_not_expected_type(value=textValue, expected_type=str, variable_name="textValue")
                        self._textValue = textValue

                    @property
                    def rawText(self) -> str:
                        return self._rawText

                    @rawText.setter
                    def rawText(self, rawText: str) -> None:
                        raise_if_variable_not_expected_type(value=rawText, expected_type=str, variable_name="rawText")
                        self._rawText = rawText

            @property
            def arguments(self) -> list:
                return self._arguments

            @arguments.setter
            def arguments(self, arguments: list) -> None:
                raise_if_variable_not_expected_type(value=arguments, expected_type=list, variable_name="arguments")
                self._arguments = arguments

        # It validate the syntax in PyCharm, but it cause a debugger error :'(
        # def __getitem__(self, i) -> InputItem:
        #    return super().__getitem__(i=i)

    @property
    def inputs(self) -> InputsCustomList:
        return self._inputs

    class Surface:
        json_key = "surface"

        def __init__(self):
            self._capabilities = list()

        @property
        def capabilities(self) -> list:
            return self._capabilities

    class AvailableSurfaces:
        json_key = "availableSurfaces"

        def __init__(self):
            self._capabilities = list()

        @property
        def capabilities(self) -> list:
            return self._capabilities

    @property
    def surface(self) -> Surface:
        return self._surface

    @property
    def availableSurfaces(self) -> AvailableSurfaces:
        return self._availableSurfaces


class OriginalDetectIntentRequest:
    json_key = "originalDetectIntentRequest"

    def __init__(self):
        self._source = str()
        self._version = str()
        self._payload = Payload()

    @property
    def payload(self) -> Payload:
        return self._payload

    @property
    def source(self) -> str:
        return self._source

    @source.setter
    def source(self, source: str) -> None:
        if not isinstance(source, str):
            raise Exception(f"source was type {type(source)} which is not valid value for his parameter.")
        self._source = source

    @property
    def version(self) -> str:
        return self._version

    @version.setter
    def version(self, version: str) -> None:
        if not isinstance(version, str):
            raise Exception(f"version was type {type(version)} which is not valid value for his parameter.")
        self._version = version

class QueryResult:
    json_key = "queryResult"

    def __init__(self):
        self._queryText = str()
        self._action = str()
        self._parameters = SafeDict()
        self._allRequiredParamsPresent = bool()
        self._fulfillmentText = str()
        self._fulfillmentMessages = list()
        self._outputContexts = list()
        self._intent = Intent()
        self._intentDetectionConfidence = int()
        self._diagnosticInfo = dict()
        self._languageCode = str()

    @property
    def queryText(self) -> str:
        return self._queryText

    @queryText.setter
    def queryText(self, queryText: str) -> None:
        if isinstance(queryText, str):
            self._queryText = queryText
        else:
            raise Exception(f"queryText was type {type(queryText)} which is not valid value for his parameter.")

    @property
    def action(self) -> str:
        return self._action

    @action.setter
    def action(self, action: str) -> None:
        if isinstance(action, str):
            self._action = action
        else:
            raise Exception(f"action was type {type(action)} which is not valid value for his parameter.")

    @property
    def parameters(self) -> SafeDict:
        if isinstance(self._parameters, dict):
            self._parameters = SafeDict(self._parameters)
        return self._parameters

    @parameters.setter
    def parameters(self, parameters: dict) -> None:
        if not isinstance(parameters, dict):
            raise Exception(f"parameters was type {type(parameters)} which is not valid value for his parameter.")
        self._parameters = SafeDict(parameters)

    @property
    def allRequiredParamsPresent(self) -> bool:
        return self._allRequiredParamsPresent

    @allRequiredParamsPresent.setter
    def allRequiredParamsPresent(self, allRequiredParamsPresent: bool) -> None:
        if isinstance(allRequiredParamsPresent, bool):
            self._allRequiredParamsPresent = allRequiredParamsPresent
        else:
            raise Exception(f"allRequiredParamsPresent was type {type(allRequiredParamsPresent)} which is not valid value for his parameter.")

    @property
    def fulfillmentText(self) -> str:
        return self._fulfillmentText

    @fulfillmentText.setter
    def fulfillmentText(self, fulfillmentText: str) -> None:
        if isinstance(fulfillmentText, str):
            self._fulfillmentText = fulfillmentText
        else:
            raise Exception(f"fulfillmentText was type {type(fulfillmentText)} which is not valid value for his parameter.")

    @property
    def fulfillmentMessages(self) -> list:
        return self._fulfillmentMessages

    @fulfillmentMessages.setter
    def fulfillmentMessages(self, fulfillmentMessages: list) -> None:
        if isinstance(fulfillmentMessages, list):
            self._fulfillmentMessages = fulfillmentMessages
        else:
            raise Exception(f"fulfillmentMessages was type {type(fulfillmentMessages)} which is not valid value for his parameter.")

    @property
    def outputContexts(self) -> list:
        return self._outputContexts

    @outputContexts.setter
    def outputContexts(self, outputContexts: list) -> None:
        if isinstance(outputContexts, list):
            self._outputContexts = outputContexts
        else:
            raise Exception(f"outputContexts was type {type(outputContexts)} which is not valid value for his parameter.")

    @property
    def intent(self) -> Intent:
        return self._intent

    @intent.setter
    def intent(self, intent: Intent) -> None:
        if isinstance(intent, Intent):
            self._intent = intent
        else:
            raise Exception(f"intent was type {type(intent)} which is not valid value for his parameter.")

    @property
    def intentDetectionConfidence(self) -> int:
        return self._intentDetectionConfidence

    @intentDetectionConfidence.setter
    def intentDetectionConfidence(self, intentDetectionConfidence: int) -> None:
        if isinstance(intentDetectionConfidence, int):
            self._intentDetectionConfidence = intentDetectionConfidence
        else:
            raise Exception(f"intentDetectionConfidence was type {type(intentDetectionConfidence)} which is not valid value for his parameter.")

    @property
    def diagnosticInfo(self) -> dict:
        return self._diagnosticInfo

    @diagnosticInfo.setter
    def diagnosticInfo(self, diagnosticInfo: dict) -> None:
        if isinstance(diagnosticInfo, dict):
            self._diagnosticInfo = diagnosticInfo
        else:
            raise Exception(f"diagnosticInfo was type {type(diagnosticInfo)} which is not valid value for his parameter.")

    @property
    def languageCode(self) -> str:
        return self._languageCode

    @languageCode.setter
    def languageCode(self, languageCode: str) -> None:
        if isinstance(languageCode, str):
            self._languageCode = languageCode
        else:
            raise Exception(f"languageCode was type {type(languageCode)} which is not valid value for his parameter.")

class Request:
    json_key = "request"

    def __init__(self):
        # General for LaunchRequest, IntentRequest and SessionEndedRequest
        self._responseId = str()
        self._queryResult = QueryResult()
        self._originalDetectIntentRequest = OriginalDetectIntentRequest()
        self._session = str()

    def process_and_set_json_to_object(self, stringed_request_json_dict: str):
        NestedObjectToDict.process_and_set_json_to_object(object_class_to_set_to=self,
                                                                  request_json_dict_stringed_dict_or_list=stringed_request_json_dict)

    def is_option_select_request(self) -> bool:
        return self.queryResult.queryText == "actions_intent_OPTION"

    def selected_option_identifier(self) -> str:
        argument_item = self.originalDetectIntentRequest.payload.get_first_input_of_type(self.originalDetectIntentRequest.payload.INPUT_TYPE_OPTION)
        if isinstance(argument_item, self.originalDetectIntentRequest.payload.InputsCustomList.InputItem.ArgumentItemsCustomList.ArgumentItem):
            return argument_item.textValue

    def is_launch_request(self) -> bool:
        if self.queryResult.intent.displayName == "LaunchIntentRequest":
            return True
        else:
            return False

    def is_in_intent_names(self, intent_names_list: list):
        if self.queryResult.intent.displayName in intent_names_list:
            return True
        return False

    def get_intent_parameter_value(self, parameter_key: str, default=None):
        return self.queryResult.parameters.get(dict_key=parameter_key).to_any(default=default)

    def is_not_usable(self):
        return False
        if self.type is not None and self.type not in [self.LaunchRequestKeyName, self.IntentRequestKeyName, self.SessionEndedRequestKeyName]:
            raise Exception(f"The request type '{self.type}' is not None or any of the supported types.")

        return False
        if (self._type == str()
        or self._requestId == str()
        or self._timestamp == str()
        or self._locale == str()):
            return True
        else:
            return False

    def to_dict(self) -> dict:
        dict_object = NestedObjectToDict.get_dict_from_nested_object(object_to_process=self,
                                                                     key_names_identifier_objects_to_go_into=["json_key"])
        return dict_object

    @property
    def queryResult(self) -> QueryResult:
        return self._queryResult

    @property
    def originalDetectIntentRequest(self) -> OriginalDetectIntentRequest:
        return self._originalDetectIntentRequest

    @property
    def responseId(self) -> str:
        return self._responseId

    @responseId.setter
    def responseId(self, responseId: str) -> None:
        if not isinstance(responseId, str):
            raise Exception(f"responseId was type {type(responseId)} which is not valid value for his parameter.")
        self._responseId = responseId

    @property
    def session(self) -> str:
        return self._session

    @session.setter
    def session(self, session: str) -> None:
        if not isinstance(session, str):
            raise Exception(f"session was type {type(session)} which is not valid value for his parameter.")
        self._session = session



# -*- coding: utf-8 -*-

import re

from inoft_vocal_framework.safe_dict import SafeDict
from inoft_vocal_framework.skill_builder.skill_settings import Settings
from inoft_vocal_framework.session_utils import add_new_played_category  # todo: remove this dependance
from inoft_vocal_framework.general_utils import pick_msg # todo: and remove this dependance too
from inoft_vocal_framework.exceptions import raise_if_variable_not_expected_type


class SpeechsList:
    def __init__(self, id: str, speechs: list = None):
        self.id = id
        self.speechs_list = speechs
        self._use_database_dynamic_messages = None

    @property
    def use_database_dynamic_messages(self):
        if self._use_database_dynamic_messages is None:
            self._use_database_dynamic_messages = Settings().settings.get("use_database_dynamic_messages").to_bool()
        return self._use_database_dynamic_messages

    @property
    def id(self) -> str:
        return self._id
    
    @id.setter
    def id(self, id: str) -> None:
        raise_if_variable_not_expected_type(value=id, expected_type=str, variable_name="id")
        self._id = id

    def speechs(self, speechs_list: list):
        self.speechs_list = speechs_list
        return self

    def pick(self, ids_messages_to_exclude=None) -> str:
        if self.use_database_dynamic_messages is True:
            from inoft_vocal_framework.databases.dynamodb.dynamodb import DynamoDbMessagesAdapter
            self.speechs_list = DynamoDbMessagesAdapter(table_name="test_messages", region_name="eu-west-3").get_speechs_list(messages_list_id=self.id)
        return pick_msg(self.speechs_list)

    def do_not_register(self):
        self.should_register_when_picking = True
        return self

    def register_played(self):
        add_new_played_category(handler_input=None, new_played_interactions_types=self.interaction_type_name)
        return self

class Speech:

    VALID_INTERPRET_AS = ('characters', 'spell-out', 'cardinal', 'number',
                          'ordinal', 'digits', 'fraction', 'unit', 'date',
                          'time', 'telephone', 'address', 'interjection', 'expletive')

    VALID_PROSODY_ATTRIBUTES = {
        'rate': ('x-slow', 'slow', 'medium', 'fast', 'x-fast'),
        'pitch': ('x-low', 'low', 'medium', 'high', 'x-high'),
        'volume': ('silent', 'x-soft', 'soft', 'medium', 'loud', 'x-loud')
    }

    VALID_VOICE_NAMES = ('Ivy', 'Joanna', 'Joey', 'Justin', 'Kendra', 'Kimberly',
                        'Matthew', 'Salli', 'Nicole', 'Russell', 'Amy', 'Brian', 'Emma',
                        'Aditi', 'Raveena', 'Hans', 'Marlene', 'Vicki', 'Conchita', 'Enrique',
                        'Carla', 'Giorgio', 'Mizuki', 'Takumi', 'Celine', 'Lea', 'Mathieu')

    VALID_EMPHASIS_LEVELS = ('strong', 'moderate', 'reduced')

    def __init__(self):
        self.probability_value = None
        self.text = ""
        self.speech = ""

    def from_dict(self, speech_safedict: SafeDict):
        self.probability_value = speech_safedict.get("probability").to_float(default=None)
        self.speech = speech_safedict.get("speech").to_str()
        return self

    def set_prob(self, probability_value: float):
        self.probability_value = probability_value
        return self

    def get_text(self):
        return self.text

    def speak(self):
        """
        <speak>
        :return:
        """
        return '<speak>{}</speak>'.format(self.speech)

    def add_text(self, value):
        """
        add text
        :return:
        """

        self.text += value
        self.speech += value
        return self

    def say_as(self, value, interpret_as, is_nested=False):
        """
        <say_as>
        :param value:
        :param interpret_as:
        :param is_nested:
        :return:
        """

        if interpret_as not in self.VALID_INTERPRET_AS:
            raise ValueError('The interpret-as provided to say_as is not valid')

        ssml = '<say-as interpret-as="{interpret_as}">' \
               '{value}</say-as>'.format(interpret_as=interpret_as, value=value)

        if is_nested:
            return ssml

        self.text += value
        self.speech += ssml
        return self

    def prosody(self, value, rate='medium', pitch='medium', volume='medium', is_nested=False):
        """
        <prosody>
        :param value:
        :param rate:
        :param pitch:
        :param volume:
        :param is_nested:
        :return:
        """

        if rate not in self.VALID_PROSODY_ATTRIBUTES['rate']:
            if re.match(r'^\d+%$', rate) is None:
                raise ValueError('The rate provided to prosody is not valid')

        if pitch not in self.VALID_PROSODY_ATTRIBUTES['pitch']:
            if re.match(r'^(\+|\-)+\d+(\.\d+)*%$', pitch) is None:
                raise ValueError('The pitch provided to prosody is not valid')

        if volume not in self.VALID_PROSODY_ATTRIBUTES['volume']:
            raise ValueError('The volume provided to prosody is not valid')

        ssml = '<prosody rate="{rate}" pitch="{pitch}" volume="{volume}">' \
               '{value}</prosody>'.format(rate=rate, pitch=pitch, volume=volume, value=value)

        if is_nested:
            return ssml

        self.text += value
        self.speech += ssml
        return self

    def sub(self, value, alias, is_nested=False):
        """
        <sub>
        :param value:
        :param alias:
        :param is_nested:
        :return:
        """

        ssml = '<sub alias="{}">{}</sub>'.format(alias, value)

        if is_nested:
            return ssml

        self.text += value
        self.speech += ssml
        return self

    def lang(self, value, lang, is_nested=False):
        """
        <lang>
        :param value:
        :param lang:
        :param is_nested:
        :return:
        """

        ssml = '<lang xml:lang="{}">{}</lang>'.format(lang, value)

        if is_nested:
            return ssml

        self.text += value
        self.speech += ssml
        return self

    def voice(self, value, name, is_nested=False):
        """
        <voice>
        :param value:
        :param name:
        :return:
        """

        if name not in self.VALID_VOICE_NAMES:
            raise ValueError('The name provided to voice is not valid')

        ssml = '<voice name="{}">{}</voice>'.format(name, value)

        if is_nested:
            return ssml

        self.text += value
        self.speech += '<voice name="{}">{}</voice>'.format(name, value)
        return self

    def pause(self, time_ms: int, is_nested=False):
        """
        <break>
        :param time_ms:
        :param is_nested:
        :return:
        """

        ssml = f"<break time='{time_ms}ms'/>"

        if is_nested:
            return ssml

        self.speech += ssml
        return self

    def whisper(self, value, is_nested=False):
        """
        :param value:
        :param is_nested:
        :return:
        """

        ssml = '<amazon:effect name="whispered">{}</amazon:effect>'.format(value)

        if is_nested:
            return ssml

        self.text += value
        self.speech += ssml
        return self

    def audio_file(self, src_url, is_nested=False):
        """
        :param src_url:
        :param is_nested:
        :return:
        """

        ssml = '<audio src="{}" />'.format(src_url)

        if is_nested:
            return ssml

        self.speech += ssml
        return self

    def emphasis(self, value, level, is_nested=False):

        if level not in self.VALID_EMPHASIS_LEVELS:
            raise ValueError('The level provided to emphasis is not valid')

        ssml = '<emphasis level="strong">{}</emphasis>'.format(value)

        if is_nested:
            return ssml

        self.text += value
        self.speech += ssml
        return self

    def p(self, value,is_nested=False):
        """
        :param value:
        :param is_nested:
        :return:
        """
        ssml = '<p>{}</p>'.format(value)

        if is_nested:
            return ssml

        self.text += value
        self.speech += ssml
        return self

    def escape(self):
        """
        escapes any special characters that will cause SSML to be invalid
        :return:
        """
        pass



ERROR_PLATFORM_NOT_SUPPORTED = '''
Platform not supported.
If you are using Alexa, you must pass the request trough the Alexa Skills Kit.
If you are using Google Assistant you must use an API gateway and create a route named /googleAssistantDialogflowV1
If you are using Bixby you must use an API gateway and create a route named /samsungBixbyV1
If you are using Siri you must use an API gateway and create a route named /appleSiriV1
As of your current framework version, no other platform is supported.
'''



class Utils:
    from boto3.dynamodb.types import TypeSerializer, TypeDeserializer
    # The serializer and deserializer can be static, even in a lambda,
    # since their classes will never change according to the user.
    _serializer = None
    _deserializer = None

    @property
    def serializer(self) -> TypeSerializer:
        if Utils._serializer is None:
            Utils._serializer = Utils.TypeSerializer()
        return Utils._serializer
    
    @property
    def deserializer(self) -> TypeDeserializer:
        if Utils._deserializer is None:
            Utils._deserializer = Utils.TypeDeserializer()
        return Utils._deserializer

    def python_to_dynamodb(self, python_object):
        dynamodb_dict = self.serializer.serialize(python_object)
        dynamodb_dict_keys = list(dynamodb_dict.keys())
        if len(dynamodb_dict_keys) == 1:
            if dynamodb_dict_keys[0] == "M":
                # With a dict, DynamoDB is not expecting the first 'M' key (stands for Map)
                return dynamodb_dict["M"]
            else:
                return dynamodb_dict
        elif len(dynamodb_dict_keys) == 0:
            return dynamodb_dict
        else:
            raise Exception(f"The number of keys of the dynamodb was superior to 1. This is NOT SUPPOSED TO HAPPEN !"
                            f"\nPlease submit an issue on the GitHub page of the framework right away (https://github.com/Robinson04/inoft_vocal_framework)."
                            f"\nThere is something really wrong on our side or on AWS side.")

    def dynamodb_to_python(self, dynamodb_object):
        if isinstance(dynamodb_object, list):
            for i, item in enumerate(dynamodb_object):
                dynamodb_object[i] = self.deserializer.deserialize(item)
        elif isinstance(dynamodb_object, dict):
            for key, item in dynamodb_object.items():
                dynamodb_object[key] = self.deserializer.deserialize(item)
            return dynamodb_object
        else:
            return self.deserializer.deserialize(dynamodb_object)



"""
This file is deprecated and shall not be used. It will soon be discarded in future versions of the framework.
"""


from inoft_vocal_framework.platforms_handlers.current_used_platform_info import CurrentPlatformData


class Card:
    def __init__(self):
        self._type = None
        self._title = str()
        self._content = str()

    @property
    def type(self):
        return self._type

    @type.setter
    def type(self, type_value) -> None:
        self._type = type_value

    @property
    def title(self):
        return self._title

    @title.setter
    def title(self, title_string: str) -> None:
        self._title = title_string

    @property
    def content(self):
        return self._content

    @content.setter
    def content(self, content_string: str) -> None:
        self._content = content_string

class OutputSpeech:
    TYPE_KEY_TEXT = "PlainText"
    TYPE_KEY_SSML = "SSML"

    def __init__(self):
        self._type = None
        self._text = str()
        self._ssml = str()

    @property
    def type(self):
        return self._type

    @property
    def text(self):
        return self._text

    @property
    def ssml(self):
        return self._ssml

    def set_text(self, text: str):
        if not isinstance(text, str):
            raise Exception(f"The following text is not a str object : {text}")
        self._type = self.TYPE_KEY_TEXT
        self._text = text
        return self

    def set_ssml(self, ssml_string: str):
        if not isinstance(ssml_string, str):
            raise Exception(f"The following ssml_string is not a str object : {ssml_string}")
        self._type = self.TYPE_KEY_SSML
        self._ssml = ssml_string
        return self

    def set_based_on_type(self, value_to_set: str, type_key: str):
        if type_key == self.TYPE_KEY_SSML:
            self.set_ssml(ssml_string=value_to_set)
        elif type_key == self.TYPE_KEY_TEXT:
            self.set_text(value_to_set)
        else:
            raise Exception(f"Type key {type_key} is not supported.")
        return self

    def get_value_to_use_based_on_set_type(self) -> str:
        if self.type == self.TYPE_KEY_SSML:
            return self.ssml
        elif self.type == self.TYPE_KEY_TEXT:
            return self.text
        else:
            raise Exception(f"The set type_key was not one of the supported type keys.")


class Reprompt:
    def __init__(self):
        self.outputSpeech = OutputSpeech()


class Response:
    def __init__(self):
        self.outputSpeech = OutputSpeech()
        self.reprompt = Reprompt()
        self.card = Card()
        self._shouldEndSession = False

    @property
    def shouldEndSession(self):
        return self._shouldEndSession

    @shouldEndSession.setter
    def shouldEndSession(self, should_end_session: bool) -> None:
        if not isinstance(should_end_session, bool):
            raise Exception(f"should_end_session must be a bool object : {should_end_session}")
        self._shouldEndSession = should_end_session

    def to_platform_dict(self) -> dict:
        from inoft_vocal_framework.platforms_handlers.handler_input import HandlerInput

        if HandlerInput.is_alexa_v1 is True:
            from inoft_vocal_framework.platforms_handlers.alexa_v1 import response
            output_response = response.Response()

            if self.outputSpeech.type == self.outputSpeech.TYPE_KEY_SSML:
                output_response.outputSpeech.set_ssml(self.outputSpeech.ssml)
            elif self.outputSpeech.type == self.outputSpeech.TYPE_KEY_TEXT:
                output_response.outputSpeech.set_text(self.outputSpeech.text)

            output_response.reprompt.outputSpeech.set_text("yaaaazazazaza")

            output_response.shouldEndSession = self.shouldEndSession


            platform_adapted_response_dict = output_response.to_dict()
            print(f"Final platform adapted on Alexa-v1 : {platform_adapted_response_dict}")
            return platform_adapted_response_dict

        elif HandlerInput.is_dialogflow_v1 is True:
            from inoft_vocal_framework.platforms_handlers.dialogflow_v1 import response
            output_response = HandlerInput.handler_input.response


            """response_item = response.SimpleResponse()
            response_item.textToSpeech = self.outputSpeech.get_value_to_use_based_on_set_type()
            response_item.displayText = "Yaaaaaazaaaa !!!!!!!"
            output_response.payload.google.richResponse.add_response_item(response_item)"""

            output_response.payload.google.expectUserResponse = True if not self.shouldEndSession else False

            output_response.payload.google.userStorage = str(HandlerInput.persistent_user_data.to_dict())

            session_user_data_context_item = response.OutputContextItem()
            for key_item_saved_data, value_item_saved_data in HandlerInput.session_user_data.to_dict().items():
                session_user_data_context_item.add_set_session_attribute(key_item_saved_data, value_item_saved_data)
            output_response.add_output_context_item(session_user_data_context_item)

            platform_adapted_response_dict = output_response.to_dict()
            print(f"Final platform adapted on Google-Assistant-v1 : {platform_adapted_response_dict}")
            return platform_adapted_response_dict
        else:
            raise Exception(f"Platform with id {CurrentPlatformData.used_platform_id} is not supported.")



class CurrentUsedPlatformInfo:
    def __init__(self):
        self._is_alexa_v1 = False
        self._is_dialogflow_v1 = False
        self._is_bixby_v1 = False

    @property
    def is_alexa_v1(self) -> bool:
        return self._is_alexa_v1

    @is_alexa_v1.setter
    def is_alexa_v1(self, is_alexa_v1: bool) -> None:
        if not isinstance(is_alexa_v1, bool):
            raise Exception(f"is_alexa_v1 was type {type(is_alexa_v1)} which is not valid value for his parameter.")
        self._is_alexa_v1 = is_alexa_v1

    @property
    def is_dialogflow_v1(self) -> bool:
        return self._is_dialogflow_v1

    @is_dialogflow_v1.setter
    def is_dialogflow_v1(self, is_dialogflow_v1: bool) -> None:
        if not isinstance(is_dialogflow_v1, bool):
            raise Exception(f"is_dialogflow_v1 was type {type(is_dialogflow_v1)} which is not valid value for his parameter.")
        self._is_dialogflow_v1 = is_dialogflow_v1

    @property
    def is_bixby_v1(self) -> bool:
        return self._is_bixby_v1

    @is_bixby_v1.setter
    def is_bixby_v1(self, is_bixby_v1: bool) -> None:
        if not isinstance(is_bixby_v1, bool):
            raise Exception(f"is_dialogflow_v1 was type {type(_is_bixby_v1)} which is not valid value for his parameter.")
        self._is_bixby_v1 = is_bixby_v1



from inoft_vocal_framework.safe_dict import SafeDict
from json import loads as json_loads


class NestedObjectToDict:
    @staticmethod
    def _remove_start_underscores_from_string_key(string_key: str) -> str:
        index_first_underscore = None
        if string_key is not None:
            for i_char in range(len(string_key)):
                if string_key[i_char] == "_":
                    if index_first_underscore is None:
                        index_first_underscore = i_char
                else:
                    if index_first_underscore is not None:
                        cleaned_string_key = string_key[:index_first_underscore] + string_key[i_char:]
                        return cleaned_string_key
                    else:
                        return string_key
        return string_key

    @staticmethod
    def _is_object_empty(class_object) -> bool:
        functions_of_object = dir(class_object)
        # We use dir and not vars, because vars do not include the functions of the class

        if "do_not_include" in functions_of_object:
            # We make sure that the do_not_include var is a function that can be called and not a shadow variable
            if class_object.do_not_include():
                return True
        return False

    @staticmethod
    def _apply_return_transformations(class_object) -> bool:
        """
        :param class_object:
        :return: bool True if the object contained the return_transformations method and it has been applied, else bool False
        """
        functions_of_object = dir(class_object)
        # We use dir and not vars, because vars do not include the functions of the class

        if "return_transformations" in functions_of_object:
            # We make sure that the return_transformations var is a function that can be called and not a shadow variable
            class_object.return_transformations()
            return True
        return False

    @staticmethod
    def _process_potential_nested_object(class_object, key_names_identifier_objects_to_go_into: list, is_first_parent_object=False) -> dict:
        for accepted_parent_key_name in key_names_identifier_objects_to_go_into:
            if hasattr(class_object, accepted_parent_key_name):
                is_class_object_empty = NestedObjectToDict._is_object_empty(class_object=class_object)

                if not is_class_object_empty:
                    NestedObjectToDict._apply_return_transformations(class_object=class_object)
                    # We only need to apply the return transformations here, since an object that could
                    # contain some return transformations will always be passed to this recurrent function.
                    # If we called the return transformations somewhere else, we could end up calling them multiple times.

                    output_dict = dict()
                    keys_to_pop_after_loop_finished = list()

                    main_vars_dict = vars(class_object)
                    if isinstance(main_vars_dict, dict):
                        for main_var_key, main_var_object in main_vars_dict.items():
                            found_accepted_key_name_in_vars_of_current_object = False

                            if isinstance(main_var_object, list):
                                for i in range(len(main_var_object)):
                                    found_accepted_key_name_in_vars_of_current_child_item_object = False

                                    for accepted_child_key_name in key_names_identifier_objects_to_go_into:
                                        if hasattr(main_var_object[i], accepted_child_key_name):
                                            # If the object has an accepted child key name, we will set the values
                                            # of the object inside of a dict with the child key name as the key dict
                                            found_accepted_key_name_in_vars_of_current_child_item_object = True

                                            nested_object_accepted_key_value = getattr(main_var_object[i].__class__, accepted_child_key_name)
                                            cleaned_key = NestedObjectToDict._remove_start_underscores_from_string_key(nested_object_accepted_key_value)

                                            item_output_dict = NestedObjectToDict._process_potential_nested_object(class_object=main_var_object[i],
                                                key_names_identifier_objects_to_go_into=key_names_identifier_objects_to_go_into)

                                            main_var_object[i] = {cleaned_key: item_output_dict} if cleaned_key is not None else item_output_dict
                                            # If the key is None, it means that the object needed to be go into to process it, but should not be
                                            # put inside a dict with its the cleaned_key has its name. This is the case for list items, for
                                            # examples for items of a carousel, they will be objects, that should be put in a list of items of
                                            # the carousel, and have their values put right away in a dict, and not inside a dict called "item"
                                            # that will contain all the values of the item.

                                    if found_accepted_key_name_in_vars_of_current_child_item_object is False:
                                        # Otherwise, we just process the value and put it back in the list, without creating a dict
                                        main_var_object[i] = NestedObjectToDict._process_potential_nested_object(class_object=main_var_object[i],
                                            key_names_identifier_objects_to_go_into=key_names_identifier_objects_to_go_into)
                            else:
                                for accepted_child_key_name in key_names_identifier_objects_to_go_into:
                                    if hasattr(main_var_object, accepted_child_key_name):
                                        found_accepted_key_name_in_vars_of_current_object = True

                                        is_main_var_object_empty = NestedObjectToDict._is_object_empty(class_object=main_var_object)
                                        # We remove an object that is defined as empty (see the called function for details)
                                        if is_main_var_object_empty:
                                            keys_to_pop_after_loop_finished.append(main_var_key)
                                        else:
                                            nested_object_accepted_key_value = getattr(main_var_object.__class__, accepted_child_key_name)
                                            cleaned_key = NestedObjectToDict._remove_start_underscores_from_string_key(nested_object_accepted_key_value)

                                            output_dict[cleaned_key] = NestedObjectToDict._process_potential_nested_object(class_object=main_var_object,
                                               key_names_identifier_objects_to_go_into=key_names_identifier_objects_to_go_into)

                                        # The loop we are currently in is just to see if the object has the attribute that tell use that it is an object
                                        # to go deeper into. If we found at least one, we do not need to look at the same object a second time.
                                        break

                            if found_accepted_key_name_in_vars_of_current_object is False and main_var_object is not None:
                                do_not_include_current_item = False

                                if isinstance(main_var_object, str):
                                    if main_var_object.replace(" ", "") == "":
                                        do_not_include_current_item = True

                                elif isinstance(main_var_object, dict) or isinstance(main_var_object, list):
                                    if not len(main_var_object) > 0:
                                        do_not_include_current_item = True
                                    else:
                                        if isinstance(main_var_object, dict):
                                            for key_item_main_var_object, value_item_main_var_object in main_var_object.items():
                                                main_var_object[key_item_main_var_object] = NestedObjectToDict._process_potential_nested_object(
                                                    class_object=value_item_main_var_object, key_names_identifier_objects_to_go_into=key_names_identifier_objects_to_go_into)
                                        elif isinstance(main_var_object, list):
                                            for i_item_main_var_object, item_main_var_object in enumerate(main_var_object):
                                                main_var_object[i_item_main_var_object] = NestedObjectToDict._process_potential_nested_object(
                                                    class_object=item_main_var_object, key_names_identifier_objects_to_go_into=key_names_identifier_objects_to_go_into)

                                if not do_not_include_current_item:
                                    cleaned_key = NestedObjectToDict._remove_start_underscores_from_string_key(main_var_key)
                                    output_dict[cleaned_key] = main_var_object

                    for key_to_pop in keys_to_pop_after_loop_finished:
                        # We pop the keys after the loop has finished, to avoid
                        # issues while modifying the dict items during the loop.
                        main_vars_dict.pop(key_to_pop)

                    if not is_first_parent_object:
                        return output_dict
                    else:
                        parent_object_accepted_key_value = getattr(class_object.__class__, accepted_parent_key_name)
                        cleaned_parent_key = NestedObjectToDict._remove_start_underscores_from_string_key(parent_object_accepted_key_value)
                        return {cleaned_parent_key: output_dict}

        # If no accepted_parent_key_name has been found, we return the object not
        # modified (to handle when we pass list or dict items trough the function.
        return class_object

    @staticmethod
    def get_dict_from_nested_object(object_to_process, key_names_identifier_objects_to_go_into: list) -> dict:
        if isinstance(key_names_identifier_objects_to_go_into, str):
            key_names_identifier_objects_to_go_into = [key_names_identifier_objects_to_go_into]

        output_dict = NestedObjectToDict._process_potential_nested_object(class_object=object_to_process,
           key_names_identifier_objects_to_go_into=key_names_identifier_objects_to_go_into, is_first_parent_object=True)
        if not isinstance(output_dict, dict):
            raise Exception(f"The following nested object converted to a dict was not of type dict but of type {type(output_dict)}")

        print(output_dict)
        return output_dict

    @staticmethod
    def _process_and_set_dict_to_object(object_class_to_set_to, dict_object: dict, key_names_identifier_objects_to_go_into: list):
        from inoft_vocal_framework.platforms_handlers.alexa_v1.context import Context

        unprocessed_vars_dict = vars(object_class_to_set_to)
        from types import MappingProxyType
        if isinstance(unprocessed_vars_dict, MappingProxyType):
            # Sometimes the vars dict is put in a MappingProxy object instead of a dict. A MappingProxy is like a dict,
            # but cannot have its values modified, so we convert it back to a dict since we will need to modify the values.
            unprocessed_vars_dict = dict(unprocessed_vars_dict)

        vars_safedict_key_processed_keys_names_with_value_unprocessed_variables_names = SafeDict()
        for key_unprocessed_var, value_var_object in unprocessed_vars_dict.items():
            found_accepted_key_name = False

            for accepted_child_key_name in key_names_identifier_objects_to_go_into:
                if hasattr(value_var_object, accepted_child_key_name):
                    # If the object contains an accepted child key name, we use it as the processed variable name. Like that, we have a
                    # total freedom to respect Python convention for the variable names, and still handle any type of object child name.
                    vars_safedict_key_processed_keys_names_with_value_unprocessed_variables_names.put(
                        dict_key=getattr(value_var_object, accepted_child_key_name), value_to_put=key_unprocessed_var)
                    found_accepted_key_name = True

            if found_accepted_key_name is False:
                # Otherwise we do a bit of processing on the variable name itself (like removing underscores) then we use it.
                processed_current_key_name = NestedObjectToDict.get_json_key_from_variable_name(variable_name=key_unprocessed_var)
                vars_safedict_key_processed_keys_names_with_value_unprocessed_variables_names.put(dict_key=processed_current_key_name,
                                                                                                  value_to_put=key_unprocessed_var)

        keys_to_pop_after_loop_finished = list()
        for key_request_element, value_request_element in dict_object.items():
            key_request_element = str(key_request_element).replace(" ", "")
            # In the received requests, there can sometimes be spaces in a key before closing it with an apostrophe.
            # Having a single empty space at the end of a key would not be the same as the key without the space.

            current_unprocessed_variable_name = vars_safedict_key_processed_keys_names_with_value_unprocessed_variables_names.get(key_request_element).to_any()
            if current_unprocessed_variable_name is not None:
                current_child_element_object = unprocessed_vars_dict[current_unprocessed_variable_name]

                found_accepted_key_name_in_vars_of_current_object = False
                if current_child_element_object is not None:
                    for accepted_child_key_name in key_names_identifier_objects_to_go_into:
                        if hasattr(current_child_element_object, accepted_child_key_name):
                            found_accepted_key_name_in_vars_of_current_object = True

                            NestedObjectToDict.process_and_set_json_to_object(
                                object_class_to_set_to=unprocessed_vars_dict[current_unprocessed_variable_name],
                                request_json_dict_stringed_dict_or_list=value_request_element,
                                key_names_identifier_objects_to_go_into=key_names_identifier_objects_to_go_into)

                # Even if the current child element object is None we will try to set a value to it, because a None value is just
                # a variable not initialized (like how often after an init). It does not means that it should not contain a value.
                if not found_accepted_key_name_in_vars_of_current_object:
                    do_not_include_current_item = False

                    if (isinstance(value_request_element, str)
                    or (type(value_request_element) == type and "str" in value_request_element.__bases__)):
                        if value_request_element.replace(" ", "") == "":
                            do_not_include_current_item = True

                    elif (isinstance(value_request_element, dict) or isinstance(value_request_element, list)
                    or (type(value_request_element) == type and ("dict" in value_request_element.__bases__
                                                                 or "list" in value_request_element.__bases__))):
                        if not len(value_request_element) > 0:
                            do_not_include_current_item = True

                    if not do_not_include_current_item:
                        # if (type(value_request_element) == type(current_child_element_object)
                        # or (((isinstance(value_request_element, float) or isinstance(value_request_element, int))
                        #     and type(current_child_element_object) in [int, float]))):
                        # For some variables, we might receive them in int or float, where the framework would store
                        # them with the other variable type. So we say that if the received variable is an int or a float,
                        # its considered compatible with the variable in the object if it is also a float or an int.

                        custom_set_from_function = getattr(current_child_element_object, "custom_set_from", None)
                        if custom_set_from_function is None:
                            unprocessed_vars_dict[current_unprocessed_variable_name] = value_request_element
                        # elif type(value_request_element) in current_child_element_object.__bases__:
                        else:
                            custom_set_from_function(value_request_element)

        for key_to_pop in keys_to_pop_after_loop_finished:
            unprocessed_vars_dict.pop(key_to_pop)

    @staticmethod
    def _process_and_set_list_to_object(object_class_to_set_to, list_object: list, key_names_identifier_objects_to_go_into: list):
        unprocessed_vars_dict = vars(object_class_to_set_to)
        from types import MappingProxyType
        if isinstance(unprocessed_vars_dict, MappingProxyType):
            # Sometimes the vars dict is put in a MappingProxy object instead of a dict. A MappingProxy is like a dict,
            # but cannot have its values modified, so we convert it back to a dict since we will need to modify the values.
            unprocessed_vars_dict = dict(unprocessed_vars_dict)

        vars_safedict_key_processed_keys_names_with_value_unprocessed_variables_names = SafeDict()
        for key_unprocessed_var in unprocessed_vars_dict.keys():
            processed_current_key_name = NestedObjectToDict.get_json_key_from_variable_name(variable_name=key_unprocessed_var)
            vars_safedict_key_processed_keys_names_with_value_unprocessed_variables_names.put(dict_key=processed_current_key_name,
                                                                                              value_to_put=key_unprocessed_var)

        for child_item in list_object:
            if isinstance(child_item, dict):
                for key_child_item, value_child_item in child_item.items():
                    current_unprocessed_variable_name = vars_safedict_key_processed_keys_names_with_value_unprocessed_variables_names.get(key_child_item).to_any()
                    if current_unprocessed_variable_name is not None:
                        current_child_element_object = unprocessed_vars_dict[current_unprocessed_variable_name]

                        if current_child_element_object is not None:
                            found_accepted_key_name_in_vars_of_current_object = False

                            for accepted_child_key_name in key_names_identifier_objects_to_go_into:
                                if hasattr(current_child_element_object, accepted_child_key_name):
                                    found_accepted_key_name_in_vars_of_current_object = True

                                    NestedObjectToDict.process_and_set_json_to_object(
                                        object_class_to_set_to=unprocessed_vars_dict[current_unprocessed_variable_name],
                                        request_json_dict_stringed_dict_or_list=value_child_item,
                                        key_names_identifier_objects_to_go_into=key_names_identifier_objects_to_go_into)

                            if not found_accepted_key_name_in_vars_of_current_object:
                                do_not_include_current_item = False

                                if (isinstance(value_child_item, str)
                                or (type(value_child_item) == type and "str" in value_child_item.__bases__)):
                                    if value_child_item.replace(" ", "") == "":
                                        do_not_include_current_item = True

                                elif (isinstance(value_child_item, dict) or isinstance(value_child_item, list)
                                or (type(value_child_item) == type and ("dict" in value_child_item.__bases__
                                                                        or "list" in value_child_item.__bases__))):
                                    if not len(value_child_item) > 0:
                                        do_not_include_current_item = True

                                if not do_not_include_current_item:
                                    custom_set_from_function = getattr(current_child_element_object, "custom_set_from", None)
                                    if custom_set_from_function is None:
                                        unprocessed_vars_dict[current_unprocessed_variable_name] = value_child_item
                                    else:
                                        custom_set_from_function(value_child_item)

            elif isinstance(child_item, list) or (type(child_item) == type and "list" in child_item.__bases__):
                NestedObjectToDict._process_and_set_list_to_object(object_class_to_set_to=child_item, list_object=child_item,
                                                                   key_names_identifier_objects_to_go_into=key_names_identifier_objects_to_go_into)

    @staticmethod
    def process_and_set_json_to_object(object_class_to_set_to, request_json_dict_stringed_dict_or_list, key_names_identifier_objects_to_go_into: list):
        if isinstance(request_json_dict_stringed_dict_or_list, dict):
            NestedObjectToDict._process_and_set_dict_to_object(object_class_to_set_to=object_class_to_set_to,
                                                               dict_object=request_json_dict_stringed_dict_or_list,
                                                               key_names_identifier_objects_to_go_into=key_names_identifier_objects_to_go_into)
        elif isinstance(request_json_dict_stringed_dict_or_list, str):
            NestedObjectToDict._process_and_set_dict_to_object(object_class_to_set_to=object_class_to_set_to,
                key_names_identifier_objects_to_go_into=key_names_identifier_objects_to_go_into,
                dict_object=NestedObjectToDict.get_dict_from_json(stringed_json_dict=request_json_dict_stringed_dict_or_list))
        elif isinstance(request_json_dict_stringed_dict_or_list, list):
            NestedObjectToDict._process_and_set_list_to_object(object_class_to_set_to=object_class_to_set_to,
                                                               list_object=request_json_dict_stringed_dict_or_list,
                                                               key_names_identifier_objects_to_go_into=key_names_identifier_objects_to_go_into)
        else:
            raise Exception(f"The request_json_dict_stringed_dict_or_list variable is of type "
                            f"{type(request_json_dict_stringed_dict_or_list)} but must of type {dict}, {str} or {list}")

    @staticmethod
    def get_dict_from_json(stringed_json_dict: str) -> dict:
        try:
            return json_loads(stringed_json_dict)
        except Exception as e:
            print(f"Warning ! The following string has tried to be converted in the get_dict_from_json function"
                  f"of the nested_object_to_dict file, but the following error occurred, returning None : {e}")
            return None

    @staticmethod
    def get_json_key_from_variable_name(variable_name: str):
        return NestedObjectToDict._remove_start_underscores_from_string_key(variable_name)





import os
from json import load as json_load
from json import loads as json_loads


class Simulator:
    def __init__(self, event_type: str):
        self.current_dir_path = os.path.dirname(os.path.abspath(__file__))
        self.request_samples_folderpath = os.path.join(self.current_dir_path, "request_samples")
        self.event_file_filepath = os.path.join(self.request_samples_folderpath, f"{event_type}.json")
        if not os.path.isfile(self.event_file_filepath):
            raise Exception(f"No event file was found at {self.event_file_filepath}. Please use a valid event_type name.")

    def get_event_and_context(self) -> (str, str):
        # For the simulator, we want the event and the context to be in the form of
        # strings, not in dicts. So we do not load them by using the json module.
        event = None
        with open(self.event_file_filepath, "r") as request_file:
            try:
                event = json_load(request_file)
            except Exception as e:
                try:
                    event = request_file.read()
                except Exception as e2:
                    print(e2)
            # todo: improve the json processing, since this way of replacing could replace some important informations in the actual request/response

        context = None

        return event, context



from inoft_vocal_framework.speechs.ssml_builder_core import SpeechsList, Speech

MSGS_WELCOME = SpeechsList(id="welcome", speechs=[
    Speech().add_text("Hello world friend !").set_prob(1),
    Speech().add_text("Hello world dude !").set_prob(1),
    Speech().add_text("Hello world love !").set_prob(0.1),
])



def get_dict_of_all_custom_defined_variables_of_class(class_object: object) -> dict:
    # From https://stackoverflow.com/questions/1398022/looping-over-all-member-variables-of-a-class-in-python
    output_vars_dict = dict()
    vars_of_object = vars(class_object)
    for key_var, value_var in vars_of_object.items():
        if not callable(getattr(class_object, key_var)) and not key_var.startswith("__"):
            output_vars_dict[key_var] = value_var
    return output_vars_dict

def is_text_ssml(text_or_ssml: str):
    if "<speak>" in text_or_ssml and "</speak>" in text_or_ssml:
        return True
    else:
        return False

def generate_uuid4() -> str:
    from uuid import uuid4 as uuid4_generator
    id_str = str(uuid4_generator())
    print(f"Generated id : {id_str}")
    return id_str

def load_yaml(filepath: str) -> dict:
    from yaml import safe_load, YAMLError
    with open(filepath, "r") as file_stream:
        try:
            return safe_load(file_stream)
        except YAMLError as error:
            raise Exception(f"The yaml file was not valid, and caused an error when loading."
                            f"Please check the file or recreate it with the cli : {error}")

def load_json(filepath: str) -> dict:
    from json import load as json_load
    with open(filepath, "r") as file_stream:
        try:
            return json_load(file_stream)
        except Exception as error:
            raise Exception(f"The json file was not valid, and caused an error when loading."
                            f"Please check the file or recreate it with the cli : {error}")



import logging
from collections import Callable

from inoft_vocal_framework.dummy_object import DummyObject
from inoft_vocal_framework.exceptions import raise_if_variable_not_expected_type
from inoft_vocal_framework.platforms_handlers.current_used_platform_info import CurrentUsedPlatformInfo
from inoft_vocal_framework.databases.dynamodb.dynamodb import DynamoDbAttributesAdapter
from inoft_vocal_framework.platforms_handlers.nested_object_to_dict import NestedObjectToDict
from inoft_vocal_framework.safe_dict import SafeDict
from inoft_vocal_framework.skill_builder.skill_settings import Settings


class HandlerInput(CurrentUsedPlatformInfo):
    def __init__(self):
        super().__init__()
        self.settings = Settings()
        self.session_users_data_safedict = self.settings.settings.get("sessions_users_data").to_safedict()
        self.sessions_users_data_disable_database = self.session_users_data_safedict.get("disable_database").to_bool()
        self.sessions_users_data_db_table_name = self.session_users_data_safedict.get("dynamodb").get("table_name").to_str()
        self.sessions_users_data_db_region_name = self.session_users_data_safedict.get("dynamodb").get("region_name").to_str()

        self._session_id = None
        self._is_invocation_new_session = None
        self._default_session_data_timeout = self.settings.settings.get("default_session_data_timeout").to_int()
        self._session_been_resumed = None
        self._simple_session_user_data = None
        self._smart_session_user_data = None
        self._persistent_user_id = None
        self._persistent_user_data = None
        self._interactivity_callback_functions = None
        self.data_for_database_has_been_modified = False

        self._is_option_select_request = None
        self._selected_option_identifier = None

        if (self.sessions_users_data_disable_database is not True
        and (not isinstance(self.sessions_users_data_db_table_name, str) or self.sessions_users_data_db_region_name is None)):
            raise Exception(f"The disable_database argument on the initialization of the InoftSkill has not been"
                            f"specified or set to True, yet the database table name or region name are missing.")
        self.dynamodb_adapter = (DynamoDbAttributesAdapter(table_name=self.sessions_users_data_db_table_name,
                                                           region_name=self.sessions_users_data_db_region_name,
                                                           primary_key_name="id", create_table=True)
                                 if self.sessions_users_data_disable_database is False else None)

        self._alexaHandlerInput, self._dialogFlowHandlerInput, self._bixbyHandlerInput = None, None, None

    @property
    def session_id(self):
        if self._session_id is None:
            if self.is_alexa_v1 is True:
                self._session_id = self.alexaHandlerInput.session.sessionId
            elif self.is_dialogflow_v1 is True:
                self._session_id = self.dialogFlowHandlerInput.session_id
            elif self.is_bixby_v1 is True:
                self._session_id = self.bixbyHandlerInput.request.sessionId

            if not isinstance(self._session_id, str):
                self._session_id = str(self._session_id)

        return self._session_id

    @property
    def is_invocation_new_session(self) -> bool:
        if self._is_invocation_new_session is None:
            if self.is_alexa_v1 is True:
                self._is_invocation_new_session = self.alexaHandlerInput.is_new_session
            elif self.is_dialogflow_v1 is True:
                self._is_invocation_new_session = self.dialogFlowHandlerInput.is_new_session
            elif self.is_bixby_v1 is True:
                raise
        return self._is_invocation_new_session

    @property
    def default_session_data_timeout(self):
        return self._default_session_data_timeout

    # todo: allow for the session timeout to be personallized and memorized for every user
    @default_session_data_timeout.setter
    def default_session_data_timeout(self, default_session_data_timeout) -> None:
        if default_session_data_timeout is not None and not isinstance(default_session_data_timeout, int):
            raise Exception(f"The default_session_data_timeout variable must be of type {None} or {int} but was : {default_session_data_timeout}")
        self._default_session_data_timeout = default_session_data_timeout

    @property
    def session_been_resumed(self) -> bool:
        if self._session_been_resumed is None:
            self._load_smart_session_user_data()
        return self._session_been_resumed

    @property
    def simple_session_user_data(self) -> SafeDict:
        if self._simple_session_user_data is None:
            if self.is_alexa_v1 is True:
                self._simple_session_user_data = self.alexaHandlerInput.session_attributes
            elif self.is_dialogflow_v1 is True:
                self._simple_session_user_data = self.dialogFlowHandlerInput.simple_session_user_data
            elif self.is_bixby_v1 is True:
                print("simple_session_user_data is not implemented for the bixby platform.")

            if not isinstance(self._simple_session_user_data, SafeDict):
                self._simple_session_user_data = SafeDict()
        return self._simple_session_user_data

    @property
    def smart_session_user_data(self) -> SafeDict:
        if self._smart_session_user_data is None:
            self._load_smart_session_user_data()
        return self._smart_session_user_data

    def _load_smart_session_user_data(self):
        # We use a separate function to load the smart session user data, and we do not include it in the property itself,
        # because this function can be called by the smart_session_user_data property or the session_been_resumed property.
        if self._smart_session_user_data is None:
            if self.sessions_users_data_disable_database is False:
                self._smart_session_user_data, self._session_been_resumed = self.dynamodb_adapter.get_smart_session_attributes(
                    user_id=self.persistent_user_id, session_id=self.session_id, timeout_seconds=self.default_session_data_timeout)

                if not isinstance(self._smart_session_user_data, SafeDict):
                    self._smart_session_user_data = SafeDict()
            else:
                self._smart_session_user_data = SafeDict()
        logging.debug(f"_smart_session_user_data = {self._smart_session_user_data}")

    @property
    def persistent_user_id(self) -> str:
        if not isinstance(self._persistent_user_id, str) or (self._persistent_user_id.replace(" ", "") == ""):
            if self.is_alexa_v1 is True:
                user_id = SafeDict(self.alexaHandlerInput.session.user).get("userId").to_str(default=None)
            elif self.is_dialogflow_v1 is True:
                user_id = self.dialogFlowHandlerInput.user_id
            elif self.is_bixby_v1 is True:
                user_id = self.bixbyHandlerInput.request.userId

            if not isinstance(user_id, str) or user_id.replace(" ", "") == "":
                from inoft_vocal_framework.utils.general import generate_uuid4
                self._persistent_user_id = generate_uuid4()
                # We need to set the persistent_user_id before memorizing it, because the memorize function will access the
                # persistent_user_data, and if the user_id is not set, we will get stuck in an infinite recursion loop
                self.persistent_memorize("userId", user_id)
                print(f"user_id {self._persistent_user_id} has been memorized in the database.")
            else:
                self._persistent_user_id = user_id
            logging.debug(f"_persistent_user_id = {self._persistent_user_id}")
        return self._persistent_user_id

    @property
    def persistent_user_data(self) -> SafeDict:
        if self._persistent_user_data is None:
            if self.sessions_users_data_disable_database is False:
                self._persistent_user_data = self.dynamodb_adapter.get_persistent_attributes(user_id=self.persistent_user_id)
            if not isinstance(self._persistent_user_data, SafeDict):
                self._persistent_user_data = SafeDict()
            logging.debug(f"_persistent_user_data = {self._persistent_user_data}")
        return self._persistent_user_data

    @property
    def interactivity_callback_functions(self) -> SafeDict:
        if self._interactivity_callback_functions is None:
            self._interactivity_callback_functions = self.smart_session_user_data.get("interactivity_callback_functions").to_safedict()
        return self._interactivity_callback_functions

    def load_event(self, event: dict) -> None:
        if self.is_alexa_v1 is True:
            from inoft_vocal_framework.platforms_handlers.alexa_v1.handler_input import AlexaHandlerInput
            self._alexaHandlerInput = AlexaHandlerInput(parent_handler_input=self)
            NestedObjectToDict.process_and_set_json_to_object(object_class_to_set_to=self.alexaHandlerInput,
                request_json_dict_stringed_dict_or_list=event, key_names_identifier_objects_to_go_into=["json_key"])

        elif self.is_dialogflow_v1 is True:
            from inoft_vocal_framework.platforms_handlers.dialogflow_v1.handler_input import DialogFlowHandlerInput
            self._dialogFlowHandlerInput = DialogFlowHandlerInput(parent_handler_input=self)
            NestedObjectToDict.process_and_set_json_to_object(object_class_to_set_to=self.dialogFlowHandlerInput.request,
                request_json_dict_stringed_dict_or_list=event, key_names_identifier_objects_to_go_into=["json_key"])

        elif self.is_bixby_v1 is True:
            from inoft_vocal_framework.platforms_handlers.samsungbixby_v1.handler_input import BixbyHandlerInput
            self._bixbyHandlerInput = BixbyHandlerInput(parent_handler_input=self)
            NestedObjectToDict.process_and_set_json_to_object(object_class_to_set_to=self.bixbyHandlerInput.request,
                request_json_dict_stringed_dict_or_list=event, key_names_identifier_objects_to_go_into=["json_key"])

    def save_callback_function_to_database(self, callback_functions_key_name: str, callback_function: Callable, identifier_key: str):
        # todo: fix bug where the identifier_key is not the right now if it has been modified because there were only 1 element
        # No matter if we already have functions for the different ids in the same key name, we remember the dict of all the
        # callback functions of the key name (will be an empty dict if was not present), then we add the callback for the
        # current specified identifier. Finally, we can memorize this new updated list.
        callback_functions_dict = self.session_remember(data_key=callback_functions_key_name, specific_object_type=dict)
        from inspect import getfile
        callback_functions_dict[identifier_key] = {"file_filepath_containing_callback": getfile(callback_function),
                                                   "callback_function_path": callback_function.__qualname__}

        self.session_memorize(callback_functions_key_name, callback_functions_dict)

    @property
    def is_option_select_request(self) -> bool:
        if self._is_option_select_request is None:
            if self.is_alexa_v1 is True:
                return False
            elif self.is_dialogflow_v1 is True:
                self._is_option_select_request = self.dialogFlowHandlerInput.is_option_select_request()
            elif self.is_bixby_v1 is True:
                return False
        return self._is_option_select_request

    @property
    def selected_option_identifier(self) -> str:
        if self.is_alexa_v1:
            raise
        elif self.is_dialogflow_v1 is True:
            self._selected_option_identifier = self.dialogFlowHandlerInput.selected_option_identifier()
        elif self.is_bixby_v1:
            raise
        return self._selected_option_identifier

    def is_launch_request(self) -> bool:
        if self.is_alexa_v1 is True:
            return self.alexaHandlerInput.is_launch_request()
        elif self.is_dialogflow_v1 is True:
            return self.dialogFlowHandlerInput.is_launch_request()
        elif self.is_bixby_v1 is True:
            return self.bixbyHandlerInput.is_launch_request()

    def is_in_intent_names(self, intent_names_list) -> bool:
        if self.is_alexa_v1 is True:
            return self.alexaHandlerInput.is_in_intent_names(intent_names_list=intent_names_list)
        elif self.is_dialogflow_v1 is True:
            return self.dialogFlowHandlerInput.is_in_intent_names(intent_names_list=intent_names_list)
        elif self.is_bixby_v1 is True:
            return self.bixbyHandlerInput.is_in_intent_names(intent_names_list=intent_names_list)

    def say(self, text_or_ssml: str) -> None:
        if self.is_alexa_v1 is True:
            self.alexaHandlerInput.say(text_or_ssml=text_or_ssml)
        elif self.is_dialogflow_v1 is True:
            self.dialogFlowHandlerInput.say(text_or_ssml=text_or_ssml)
        elif self.is_bixby_v1 is True:
            self.bixbyHandlerInput.say(text_or_ssml=text_or_ssml)

    def reprompt(self, text_or_ssml: str) -> None:
        if self.is_alexa_v1 is True:
            self.alexaHandlerInput.reprompt(text_or_ssml=text_or_ssml)
        elif self.is_dialogflow_v1 is True:
            self.dialogFlowHandlerInput.reprompt(text_or_ssml=text_or_ssml)
        elif self.is_bixby_v1 is True:
            self.bixbyHandlerInput.reprompt(text_or_ssml=text_or_ssml)

    def end_session(self, should_end: bool = True) -> None:
        if self.is_alexa_v1 is True:
            self.alexaHandlerInput.end_session(should_end=should_end)
        elif self.is_dialogflow_v1 is True:
            raise
        elif self.is_bixby_v1 is True:
            raise

    def get_intent_arg_value(self, arg_key: str):
        if self.is_alexa_v1 is True:
            return self.alexaHandlerInput.request.get_intent_slot_value(slot_key=arg_key)
        elif self.is_dialogflow_v1 is True:
            return self.dialogFlowHandlerInput.request.get_intent_parameter_value(parameter_key=arg_key)
        elif self.is_bixby_v1 is True:
            return self.bixbyHandlerInput.request.get_intent_parameter_value(parameter_key=arg_key)

    def simple_session_memorize(self, data_key: str, data_value=None) -> None:
        if data_value is not None and isinstance(data_key, str) and data_key != "":
            self.simple_session_user_data.put(dict_key=data_key, value_to_put=data_value)

    def simple_session_batch_memorize(self, data_dict: dict) -> None:
        if not isinstance(data_dict, dict):
            raise Exception(f"The data_dict must be of type dict but was of type {type(data_dict)}")
        else:
            for key_item, value_item in data_dict.items():
                self.simple_session_user_data.put(dict_key=key_item, value_to_put=value_item)

    def simple_session_remember(self, data_key: str, specific_object_type=None):
        data_object = self.simple_session_user_data.get(data_key)
        if specific_object_type is None:
            return data_object.to_any()
        else:
            return data_object.to_specific_type(type_to_return=specific_object_type)

    def simple_session_forget(self, data_key: str) -> None:
        self.simple_session_user_data.pop(dict_key=data_key)

    def session_memorize(self, data_key: str, data_value=None) -> None:
        if data_value is not None and isinstance(data_key, str) and data_key != "":
            self.data_for_database_has_been_modified = True
            self.smart_session_user_data.put(dict_key=data_key, value_to_put=data_value)

    def session_batch_memorize(self, data_dict: dict) -> None:
        if not isinstance(data_dict, dict):
            raise Exception(f"The data_dict must be of type dict but was of type {type(data_dict)}")
        else:
            if len(data_dict) > 0:
                self.data_for_database_has_been_modified = True

            for key_item, value_item in data_dict.items():
                self.smart_session_user_data.put(dict_key=key_item, value_to_put=value_item)

    def session_remember(self, data_key: str, specific_object_type=None):
        data_object = self.smart_session_user_data.get(data_key)
        if specific_object_type is None:
            return data_object.to_any()
        else:
            return data_object.to_specific_type(type_to_return=specific_object_type)

    def session_forget(self, data_key: str) -> None:
        self.data_for_database_has_been_modified = True
        self.smart_session_user_data.pop(dict_key=data_key)

    def persistent_memorize(self, data_key: str, data_value=None) -> None:
        if data_value is not None and isinstance(data_key, str) and data_key != "":
            self.data_for_database_has_been_modified = True
            self.persistent_user_data.put(dict_key=data_key, value_to_put=data_value)

    def persistent_batch_memorize(self, data_dict: dict) -> None:
        if not isinstance(data_dict, dict):
            raise Exception(f"The data_dict must be of type dict but was of type {type(data_dict)}")
        else:
            for key_item, value_item in data_dict.items():
                self.data_for_database_has_been_modified = True
                self.persistent_user_data.put(dict_key=key_item, value_to_put=value_item)

    def persistent_remember(self, data_key: str, specific_object_type=None):
        data_object = self.persistent_user_data.get(data_key)
        if specific_object_type is None:
            return data_object.to_any()
        else:
            return data_object.to_specific_type(type_to_return=specific_object_type)

    def persistent_forget(self, data_key: str) -> None:
        self.data_for_database_has_been_modified = True
        self.persistent_user_data.pop(dict_key=data_key)

    def memorize_session_then_state(self, state_handler_class_type_or_name) -> None:
        from inoft_vocal_framework.skill_builder.inoft_skill_builder import InoftStateHandler

        if state_handler_class_type_or_name is not None:
            then_state_class_name = None

            if isinstance(state_handler_class_type_or_name, type):
                try:
                    if InoftStateHandler in state_handler_class_type_or_name.__bases__:
                        then_state_class_name = state_handler_class_type_or_name.__name__
                    else:
                        raise Exception(f"The state handler {state_handler_class_type_or_name} did not had {InoftStateHandler} in its bases classes.")
                except Exception as e:
                    raise Exception(f"Error while setting the following session_then_state {state_handler_class_type_or_name}."
                                    f"Make sure its a class with {InoftStateHandler} as its/one of its parent class."
                                    f"No checks are being made on the class, only a try and except that returned : {e}")

            if isinstance(state_handler_class_type_or_name, str):
                then_state_class_name = state_handler_class_type_or_name

            if then_state_class_name is not None:
                self.session_memorize(data_key="thenState", data_value=then_state_class_name)
        else:
            raise Exception(f"state_handler_class_type_or_name must be an class type or str but was {state_handler_class_type_or_name}")

    def remember_session_then_state(self):
        last_session_then_state = self.session_remember("thenState", str)
        if last_session_then_state.replace(" ", "") != "":
            return last_session_then_state
        else:
            return None

    def forget_session_then_state(self) -> None:
        self.session_forget("thenState")

    def memorize_session_last_intent_handler(self, handler_class_type_instance_name) -> None:
        from inoft_vocal_framework.skill_builder.inoft_skill_builder import InoftRequestHandler, InoftStateHandler

        if handler_class_type_instance_name is not None:
            handler_class_name = None

            if isinstance(handler_class_type_instance_name, str):
                handler_class_name = handler_class_type_instance_name
            else:
                try:
                    handler_class_type_object = handler_class_type_instance_name if callable(handler_class_type_instance_name) else handler_class_type_instance_name.__class__
                    if InoftRequestHandler in handler_class_type_object.__bases__ or InoftStateHandler in handler_class_type_object.__bases__:
                        handler_class_name = handler_class_type_object.__name__
                    else:
                        raise Exception(f"The state handler {handler_class_type_object} did not had {InoftStateHandler} in its bases classes.")
                except Exception as e:
                    raise Exception(f"Error while setting the following session_last_intent_handler {handler_class_type_instance_name}."
                                    f"Make sure its a class with {InoftStateHandler} as its/one of its parent class."
                                    f"No checks are being made on the class, only a try and except that returned : {e}")

            if handler_class_name is not None:
                self.session_memorize(data_key="lastIntentHandler", data_value=handler_class_name)
                print("SAVED")
        else:
            raise Exception(f"handler_class_type_or_name must be an class type or str but was {handler_class_type_instance_name}")

    def remember_session_last_intent_handler(self):
        last_intent_handler_class_name = self.session_remember("lastIntentHandler", str)
        if last_intent_handler_class_name.replace(" ", "") != "":
            return last_intent_handler_class_name
        else:
            return None

    def forget_session_last_intent_handler(self) -> None:
        self.session_forget("lastIntentHandler")

    def save_attributes_if_need_to(self):
        if self.data_for_database_has_been_modified:
            if self.sessions_users_data_disable_database is False:
                self.dynamodb_adapter.save_attributes(user_id=self.persistent_user_id, session_id=self.session_id,
                                                      smart_session_attributes=self.smart_session_user_data.to_dict(),
                                                      persistent_attributes=self.persistent_user_data.to_dict())

    def to_platform_dict(self) -> dict:
        output_response_dict = None
        # todo: improve this code, i found it dirty...
        if self.is_alexa_v1 is True:
            output_response_dict = {
                "version": "1.0",
                "sessionAttributes": self.simple_session_user_data.to_dict(),
                "response": self.alexaHandlerInput.response.to_dict()["response"]  # todo: fix the need to enter with response key
            }
        elif self.is_dialogflow_v1 is True:
            self.dialogFlowHandlerInput.response.payload.google.userStorage = str({"userId": self.persistent_user_id})

            from inoft_vocal_framework.platforms_handlers.dialogflow_v1.response import OutputContextItem
            session_user_data_context_item = OutputContextItem(session_id=self.dialogFlowHandlerInput.session_id,
                                                               name=OutputContextItem.session_data_name)

            for key_item_saved_data, value_item_saved_data in self.simple_session_user_data.to_dict().items():
                session_user_data_context_item.add_set_session_attribute(key_item_saved_data, value_item_saved_data)
            self.dialogFlowHandlerInput.response.add_output_context_item(session_user_data_context_item)

            output_response_dict = self.dialogFlowHandlerInput.response.to_dict()
        elif self.is_bixby_v1 is True:
            output_response_dict = self.bixbyHandlerInput.response.to_dict()

        return output_response_dict

    @property
    def alexaHandlerInput(self):
        return self._alexaHandlerInput if self._alexaHandlerInput is not None else DummyObject()

    @property
    def alexa(self):
        return self.alexaHandlerInput

    @property
    def dialogFlowHandlerInput(self):
        return self._dialogFlowHandlerInput if self._dialogFlowHandlerInput is not None else DummyObject()

    @property
    def google(self):
        return self.dialogFlowHandlerInput

    @property
    def bixbyHandlerInput(self):
        return self._bixbyHandlerInput if self._bixbyHandlerInput is not None else DummyObject()

    @property
    def bixby(self):
        return self.bixbyHandlerInput


class HandlerInputWrapper:
    from inoft_vocal_framework.platforms_handlers.dialogflow_v1.handler_input import DialogFlowHandlerInput
    from inoft_vocal_framework.platforms_handlers.alexa_v1.handler_input import AlexaHandlerInput
    from inoft_vocal_framework.platforms_handlers.samsungbixby_v1.handler_input import BixbyHandlerInput

    def __init__(self, parent_handler=None):
        if parent_handler is None:
            self._handler_input = None
        else:
            try:
                self._handler_input = parent_handler.handler_input
            except Exception as e:
                raise Exception(f"A parent_handler has been specified, but do not has an handler_input property : {parent_handler} : {e}")

    @property
    def handler_input(self) -> HandlerInput:
        return self._handler_input

    @handler_input.setter
    def handler_input(self, handler_input: HandlerInput) -> None:
        if not isinstance(handler_input, HandlerInput):
            raise Exception(f"handler_input was type {type(handler_input)} which is not valid value for his parameter.")
        self._handler_input = handler_input

    @property
    def session_been_resumed(self):
        return self.handler_input.session_been_resumed

    def is_launch_request(self) -> bool:
        return self.handler_input.is_launch_request()

    def is_in_intent_names(self, intent_names_list) -> bool:
        return self.handler_input.is_in_intent_names(intent_names_list=intent_names_list)

    def say(self, text_or_ssml: str):
        self.handler_input.say(text_or_ssml=text_or_ssml)
        return self

    def reprompt(self, text_or_ssml: str):
        self.handler_input.reprompt(text_or_ssml=text_or_ssml)
        return self

    def get_intent_arg_value(self, arg_key: str) -> SafeDict:
        return self.handler_input.get_intent_arg_value(arg_key=arg_key)

    def simple_session_memorize(self, data_key: str, data_value=None):
        self.handler_input.simple_session_memorize(data_key=data_key, data_value=data_value)
        return self

    def end_session(self, should_end: bool = True):
        self.handler_input.end_session(should_end=should_end)
        return self

    def simple_session_batch_memorize(self, data_dict: dict):
        self.handler_input.simple_session_batch_memorize(data_dict=data_dict)
        return self

    def simple_session_remember(self, data_key: str, specific_object_type=None):
        return self.handler_input.simple_session_remember(data_key=data_key, specific_object_type=specific_object_type)

    def simple_session_forget(self, data_key: str):
        self.handler_input.simple_session_forget(data_key=data_key)
        return self

    def session_memorize(self, data_key: str, data_value=None):
        self.handler_input.session_memorize(data_key=data_key, data_value=data_value)
        return self

    def session_batch_memorize(self, data_dict: dict):
        self.handler_input.session_batch_memorize(data_dict=data_dict)
        return self

    def session_remember(self, data_key: str, specific_object_type=None):
        return self.handler_input.session_remember(data_key=data_key, specific_object_type=specific_object_type)

    def session_forget(self, data_key: str):
        self.handler_input.session_forget(data_key=data_key)
        return self

    def persistent_memorize(self, data_key: str, data_value=None):
        self.handler_input.persistent_memorize(data_key=data_key, data_value=data_value)
        return self

    def persistent_batch_memorize(self, data_dict: dict):
        self.handler_input.persistent_batch_memorize(data_dict=data_dict)
        return self

    def persistent_remember(self, data_key: str, specific_object_type=None):
        return self.handler_input.persistent_remember(data_key=data_key, specific_object_type=specific_object_type)

    def persistent_forget(self, data_key: str):
        self.handler_input.persistent_forget(data_key=data_key)
        return self

    def memorize_session_then_state(self, state_handler_class_type_or_name):
        self.handler_input.memorize_session_then_state(state_handler_class_type_or_name=state_handler_class_type_or_name)
        return self

    def remember_session_then_state(self):
        self.handler_input.remember_session_then_state()
        return self

    def forget_session_then_state(self):
        self.handler_input.forget_session_then_state()

    # todo: add the then states saved in the persistent attributes, allow to
    #  set who should take over if there is a then state in persistent and session

    def to_platform_dict(self) -> dict:
        return self.handler_input.to_platform_dict()

    @property
    def is_alexa_v1(self) -> bool:
        return self.handler_input.is_alexa_v1

    @property
    def is_dialogflow_v1(self) -> bool:
        return self.handler_input.is_dialogflow_v1

    @property
    def is_bixby_v1(self) -> bool:
        return self.handler_input.is_bixby_v1

    @property
    def alexa(self) -> AlexaHandlerInput:
        return self.handler_input.alexaHandlerInput

    @property
    def google(self) -> DialogFlowHandlerInput:
        return self.handler_input.dialogFlowHandlerInput

    @property
    def bixby(self) -> BixbyHandlerInput:
        return self.handler_input.bixbyHandlerInput



import logging

from inoft_vocal_framework.dummy_object import DummyObject
from inoft_vocal_framework.exceptions import raise_if_variable_not_expected_type, raise_if_value_not_in_list, raise_if_variable_not_expected_type_and_not_none
from inoft_vocal_framework.platforms_handlers.alexa_v1.response.response import Response
from inoft_vocal_framework.safe_dict import SafeDict


class AudioPlayer:
    json_key = None
    TYPE_PLAY = "AudioPlayer.Play"
    TYPE_STOP = "AudioPlayer.Stop"
    TYPE_CLEAR_QUEUE = "AudioPlayer.ClearQueue"
    AVAILABLE_TYPES = [TYPE_PLAY, TYPE_STOP, TYPE_CLEAR_QUEUE]
    PLAY_BEHAVIOR_REPLACE_ALL = "REPLACE_ALL"
    AVAILABLE_PLAY_BEHAVIORS = [PLAY_BEHAVIOR_REPLACE_ALL]
    CLEAR_BEHAVIOR_CLEAR_QUEUE = "CLEAR_ENQUEUED"
    CLEAR_BEHAVIOR_CLEAR_ALL = "CLEAR_ALL"
    AVAILABLE_CLEAR_BEHAVIORS = [CLEAR_BEHAVIOR_CLEAR_QUEUE, CLEAR_BEHAVIOR_CLEAR_ALL]

    # todo: include a warning when using an audio played if the skill do not has the
    #  audio player setting activated (i lost like 20 minutes on this dumb thing)

    class AudioItem:
        json_key = "audioItem"

        class Stream:
            json_key = "stream"

            def __init__(self, token_identifier: str, url: str, offsetInMilliseconds: int = 0):
                self.token = token_identifier
                self.url = url
                self.offsetInMilliseconds = offsetInMilliseconds

            @property
            def token(self) -> str:
                return self._token

            @token.setter
            def token(self, token: str) -> None:
                raise_if_variable_not_expected_type(value=token, expected_type=str, variable_name="token")
                self._token = token

            @property
            def url(self) -> str:
                return self._url

            @url.setter
            def url(self, url: str) -> None:
                raise_if_variable_not_expected_type(value=url, expected_type=str, variable_name="url")
                self._url = url

            @property
            def offsetInMilliseconds(self) -> int:
                return self._offsetInMilliseconds

            @offsetInMilliseconds.setter
            def offsetInMilliseconds(self, offsetInMilliseconds: int) -> None:
                raise_if_variable_not_expected_type(value=offsetInMilliseconds, expected_type=int,
                                                    variable_name="offsetInMilliseconds")
                self._offsetInMilliseconds = offsetInMilliseconds

        def __init__(self, token_identifier: str, url: str, offsetInMilliseconds: int = 0):
            self.stream = self.Stream(token_identifier=token_identifier, url=url,
                                      offsetInMilliseconds=offsetInMilliseconds)

    def __init__(self, played_type: str, play_behavior: str = None, clear_behavior: str = None,
                 token_identifier: str = None, url: str = None, offsetInMilliseconds: int = 0):
        self.type = played_type
        self._playBehavior = play_behavior
        self._clearBehavior = clear_behavior

        if token_identifier is not None and url is not None:
            self.audioItem = self.AudioItem(token_identifier=token_identifier, url=url,
                                            offsetInMilliseconds=offsetInMilliseconds)

    @property
    def type(self) -> str:
        return self._type

    @type.setter
    def type(self, type_value: str) -> None:
        raise_if_value_not_in_list(value=type_value, list_object=self.AVAILABLE_TYPES, variable_name="type")
        self._type = type_value

    @property
    def playBehavior(self) -> str:
        return self._playBehavior

    @playBehavior.setter
    def playBehavior(self, playBehavior: str) -> None:
        raise_if_value_not_in_list(value=playBehavior, list_object=self.AVAILABLE_PLAY_BEHAVIORS, variable_name="playBehavior")
        self._playBehavior = playBehavior

    @property
    def clearBehavior(self) -> str:
        return self._clearBehavior

    @clearBehavior.setter
    def clearBehavior(self, clearBehavior: str) -> None:
        raise_if_value_not_in_list(value=clearBehavior, list_object=self.AVAILABLE_CLEAR_BEHAVIORS, variable_name="clearBehavior")
        self._clearBehavior = clearBehavior


class AudioItemObject(object):
    # This object is not used for parsing, but for data manipulation
    def __init__(self, identifier: str, audio_item_dict: dict = None):
        self.identifier = identifier

        if audio_item_dict is not None:
            if isinstance(audio_item_dict, dict):
                audio_item_dict = SafeDict(audio_item_dict)
            if not isinstance(audio_item_dict, SafeDict):
                raise Exception(f"The audio_item_dict must be of type dict or SafeDict but was {type(audio_item_dict)}: {audio_item_dict}")

            self.mp3_file_url = audio_item_dict.get("url").to_str()
            self.title = audio_item_dict.get("title").to_str()
            self.subtitle = audio_item_dict.get("subtitle").to_str()
            self.offset_in_milliseconds = audio_item_dict.get("offsetInMilliseconds").to_int()
            self.icon_image_url = audio_item_dict.get("iconImageUrl").to_str()
            self.background_image_url = audio_item_dict.get("backgroundImageUrl").to_str()

    def to_dict(self):
        return {
            "identifier": self.identifier,
            "url": self.mp3_file_url,
            "title": self.title,
            "subtitle": self.subtitle,
            "offsetInMilliseconds": self.offset_in_milliseconds,
            "iconImageUrl": self.icon_image_url,
            "backgroundImageUrl": self.background_image_url,
        }

    @property
    def identifier(self) -> str:
        return self._identifier

    @identifier.setter
    def identifier(self, identifier: str) -> None:
        raise_if_variable_not_expected_type(value=identifier, expected_type=str, variable_name="identifier")
        self._identifier = identifier

    @property
    def mp3_file_url(self) -> str:
        return self._mp3_file_url

    @mp3_file_url.setter
    def mp3_file_url(self, mp3_file_url: str) -> None:
        raise_if_variable_not_expected_type_and_not_none(value=mp3_file_url, expected_type=str, variable_name="mp3_file_url")
        self._mp3_file_url = mp3_file_url

    @property
    def title(self) -> str:
        return self._title

    @title.setter
    def title(self, title: str) -> None:
        raise_if_variable_not_expected_type_and_not_none(value=title, expected_type=str, variable_name="title")
        self._title = title

    @property
    def subtitle(self) -> str:
        return self._subtitle

    @subtitle.setter
    def subtitle(self, subtitle: str) -> None:
        raise_if_variable_not_expected_type_and_not_none(value=subtitle, expected_type=str, variable_name="subtitle")
        self._subtitle = subtitle

    @property
    def offset_in_milliseconds(self) -> int:
        return self._offset_in_milliseconds

    @offset_in_milliseconds.setter
    def offset_in_milliseconds(self, offset_in_milliseconds: int) -> None:
        if isinstance(offset_in_milliseconds, str) and offset_in_milliseconds.isdigit():
            offset_in_milliseconds = int(offset_in_milliseconds)
        raise_if_variable_not_expected_type_and_not_none(value=offset_in_milliseconds, expected_type=int, variable_name="offset_in_milliseconds")
        self._offset_in_milliseconds = offset_in_milliseconds

    @property
    def icon_image_url(self) -> str:
        return self._icon_image_url

    @icon_image_url.setter
    def icon_image_url(self, icon_image_url: str) -> None:
        raise_if_variable_not_expected_type_and_not_none(value=icon_image_url, expected_type=str, variable_name="icon_image_url")
        self._icon_image_url = icon_image_url

    @property
    def background_image_url(self) -> str:
        return self._background_image_url

    @background_image_url.setter
    def background_image_url(self, background_image_url: str) -> None:
        raise_if_variable_not_expected_type_and_not_none(value=background_image_url, expected_type=str, variable_name="background_image_url")
        self._background_image_url = background_image_url

class AudioPlayerWrapper:
    def __init__(self, parent_handler_input):
        self.parent_handler_input = parent_handler_input
        self._audioPlayer = None

        self._history = None

    @property
    def audioPlayer(self):
        if self._audioPlayer is None:
            return DummyObject()
        return self._audioPlayer

    @audioPlayer.setter
    def audioPlayer(self, audioPlayer: AudioPlayer) -> None:
        if self._audioPlayer is not None:
            logging.warning(f"The AudioPlayer has been with a playBehavior of {audioPlayer.playBehavior} where it had already"
                            f"been set to another AudioPlayer with a playBehavior of {self._audioPlayer.playBehavior}."
                            f"Normally you should set the AudioPlayer a maximum of one time per invocation, since if"
                            f"you set it again you override the previously set one.")

        raise_if_variable_not_expected_type(value=audioPlayer, expected_type=AudioPlayer, variable_name="audioPlayer")
        self._audioPlayer = audioPlayer
        self.parent_handler_input.alexa.response.directives.append(audioPlayer)

        from inoft_vocal_framework.platforms_handlers.alexa_v1.audioplayer.audioplayer_handlers import AlexaAudioPlayerHandlers
        self.parent_handler_input.alexa.save_audioplayer_handlers_group_class(handlers_group_class_type=AlexaAudioPlayerHandlers)

    def play(self, identifier: str, mp3_file_url: str, title: str, subtitle: str = None, icon_image_url: str = None, background_image_url: str = None,
             offset_in_milliseconds: int = 0, play_behavior: str = AudioPlayer.PLAY_BEHAVIOR_REPLACE_ALL, should_end_session: bool = True):
        # todo: check validity of mp3_file_url

        audio_item = AudioItemObject(identifier=identifier)
        audio_item.mp3_file_url = mp3_file_url
        audio_item.title = title
        audio_item.subtitle = subtitle
        audio_item.icon_image_url = icon_image_url
        audio_item.background_image_url = background_image_url
        audio_item.offset_in_milliseconds = offset_in_milliseconds

        self.audioPlayer = AudioPlayer(played_type=AudioPlayer.TYPE_PLAY, play_behavior=play_behavior,
                                       token_identifier=audio_item.identifier, url=audio_item.mp3_file_url,
                                       offsetInMilliseconds=audio_item.offset_in_milliseconds)

        self.memorize_audio_item(audio_item=audio_item)
        self.id_last_played_item = audio_item.identifier
        self.parent_handler_input.alexa.end_session(should_end=should_end_session)

    def play_item(self, audio_item: AudioItemObject, play_behavior: str = AudioPlayer.PLAY_BEHAVIOR_REPLACE_ALL, should_end_session: bool = True):
        # todo: check validity of mp3_file_url

        self.audioPlayer = AudioPlayer(played_type=AudioPlayer.TYPE_PLAY, play_behavior=play_behavior,
                                       token_identifier=audio_item.identifier, url=audio_item.mp3_file_url,
                                       offsetInMilliseconds=audio_item.offset_in_milliseconds)

        self.memorize_audio_item(audio_item=audio_item)
        self.id_last_played_item = audio_item.identifier
        self.parent_handler_input.alexa.end_session(should_end=should_end_session)

    def stop(self, should_end_session: bool = False):
        self.audioPlayer = AudioPlayer(played_type=AudioPlayer.TYPE_STOP)
        # todo: remember in database the offset in milliseconds
        self.parent_handler_input.alexa.end_session(should_end=should_end_session)

    def clear_queue(self, also_stop_and_clear_current_played_media: bool = False, should_end_session: bool = False):
        # todo: remember in database that we clear the queue (when the queue will be integrated)
        self.audioPlayer = AudioPlayer(played_type=AudioPlayer.TYPE_CLEAR_QUEUE,
                                       clear_behavior=(AudioPlayer.CLEAR_BEHAVIOR_CLEAR_QUEUE
                                                       if also_stop_and_clear_current_played_media is False
                                                       else AudioPlayer.CLEAR_BEHAVIOR_CLEAR_ALL))
        self.parent_handler_input.alexa.end_session(should_end=should_end_session)

    @property
    def token(self) -> str:
        return self.parent_handler_input.alexa.context.audioPlayer.token

    @property
    def player_activity(self) -> str:
        return self.parent_handler_input.alexa.context.audioPlayer.playerActivity

    @property
    def offset_in_milliseconds(self) -> str:
        return self.parent_handler_input.alexa.context.audioPlayer.offsetInMilliseconds

    @property
    def history(self) -> SafeDict:
        if self._history is None:
            self._history = self.parent_handler_input.persistent_remember(data_key="audioPlayerHistory", specific_object_type=SafeDict)
        return self._history

    def get_last_played_item(self) -> AudioItemObject:
        id_last_played_item = self.id_last_played_item
        if id_last_played_item is not None:
            audio_item_dict = self.history.get("items").get(id_last_played_item).to_dict(default=None)
            if audio_item_dict is not None:
                audio_item_object = AudioItemObject(identifier=id_last_played_item, audio_item_dict=audio_item_dict)
                return audio_item_object
        return None

    @property
    def id_last_played_item(self):
        return self.history.get("idLastPlayedItem").to_str(default=None)

    @id_last_played_item.setter
    def id_last_played_item(self, id_last_played_item: str) -> None:
        raise_if_variable_not_expected_type(value=id_last_played_item, expected_type=str, variable_name="id_last_played_item")
        self.history.put("idLastPlayedItem", id_last_played_item)
        self.parent_handler_input.persistent_memorize(data_key="audioPlayerHistory", data_value=self.history)

    @property
    def history_items_dicts(self) -> list:
        return self.history.get("items").to_list()

    def get_history_items_objects(self) -> list:
        output_items_objects_list = list()
        history_items_dict = self.history_items_dicts
        for i, item in enumerate(history_items_dict):
            output_items_objects_list.append(AudioItemObject(identifier=str(i), audio_item_dict=item))
        return output_items_objects_list

    def memorize_audio_item(self, audio_item: AudioItemObject) -> None:
        self.history.get_set("items", {}).put(audio_item.identifier, audio_item.to_dict())
        self.parent_handler_input.persistent_memorize(data_key="audioPlayerHistory", data_value=self.history)








import os
from pathlib import Path
from tkinter import Tk
from tkinter import PhotoImage

import inoft_vocal_framework


def _make_tk_root() -> Tk:
    root = Tk()
    root.withdraw()

    logo_filepath = os.path.join(str(Path(os.path.dirname(os.path.abspath(inoft_vocal_framework.__file__))).parent), "img", "inoft_small_logo.png")
    icon = PhotoImage(master=root, file=logo_filepath)
    root.wm_iconphoto(True, icon)

    return root

def select_folder(title: str = None, initial_dir: str = None):
    _make_tk_root()
    from tkinter.filedialog import askdirectory

    kwargs = {"title": title or "Select a folder", "mustexist": True}
    if initial_dir is not None and os.path.isdir(initial_dir):
        kwargs["initialdir"] = initial_dir

    while True:
        folderpath = askdirectory(**kwargs)
        if os.path.isdir(folderpath) and False:
            return folderpath
        else:
            from tkinter import messagebox
            ok = messagebox.askokcancel(title="Invalid folder", message=f"Folderpath {folderpath} was not valid. Please select a new one or quit.")
            if ok is not True:
                return None



from setuptools import setup

setup(
    name="inoftvocal",
    version="0.66.3",
    packages=["inoft_vocal_framework", "inoft_vocal_framework.cli", "inoft_vocal_framework.cli.core",
              "inoft_vocal_framework.utils", "inoft_vocal_framework.templates",
              "inoft_vocal_framework.templates.hello_world_1", "inoft_vocal_framework.templates.simple_adventure_game_1",
              "inoft_vocal_framework.speechs", "inoft_vocal_framework.databases",
              "inoft_vocal_framework.databases.dynamodb", "inoft_vocal_framework.skill_builder",
              "inoft_vocal_framework.platforms_handlers", "inoft_vocal_framework.platforms_handlers.alexa_v1",
              "inoft_vocal_framework.platforms_handlers.alexa_v1.response",
              "inoft_vocal_framework.platforms_handlers.alexa_v1.audioplayer",
              "inoft_vocal_framework.platforms_handlers.simulator",
              "inoft_vocal_framework.platforms_handlers.simulator.request_samples",
              "inoft_vocal_framework.platforms_handlers.simulator.request_samples.alexa",
              "inoft_vocal_framework.platforms_handlers.simulator.request_samples.google",
              "inoft_vocal_framework.platforms_handlers.dialogflow_v1",
              "inoft_vocal_framework.platforms_handlers.samsungbixby_v1",
              "inoft_vocal_framework.platforms_handlers.siri",
              "inoft_vocal_framework.platforms_handlers.endpoints_providers"],
    include_package_data=True,
    install_requires=["PyYAML", "cerberus", "boto3", "click"],
    entry_points={
        "console_scripts": [
            "inoft = inoft_vocal_framework.cli.cli_index:cli",
            "inoftvocal = inoft_vocal_framework.cli.cli_index:cli",
        ],
    },
    url="https://github.com/Robinson04/inoft_vocal_framework",
    download_url="https://github.com/Robinson04/inoft_vocal_framework/archive/0.62.3.tar.gz",
    license="MIT",
    author="Inoft",
    author_email="robinson@inoft.com",
    description="Create advanced cross-platform skills for Alexa, Google Assistant and Samsung Bixby",
)




from concurrent.futures import ThreadPoolExecutor
import time
import boto3
from inoft_vocal_framework.cli.core.policies import ATTACH_POLICY, ASSUME_POLICY


class CoreClients:
    def __init__(self):  #, pool: ThreadPoolExecutor):
        self.attach_policy = ATTACH_POLICY
        self.assume_policy = ASSUME_POLICY

        self.boto_session = None
        self.lambda_client = None
        self.s3_client = None
        self.dynamodb_client = None
        self.api_gateway_client = None
        self.events_client = None

        # east_config = boto3.session.Config(region_name="us-east-1")
        # self.acm_client = boto3.client("acm", config=east_config)
        self.iam_client = None
        self.iam_resource = None
        # self.iam = self.boto_session.client("iam")
        # self.cloudformation_client = boto3.client("cloudformation")
        # self.cloudwatch = boto3.client("cloudwatch")

        # self.pool = pool
        self.pool = ThreadPoolExecutor(7)
        # We cannot use asyncio for the initialization of the client, because currently the http
        # requests made by boto3 are blocking http request made with the default urllib library.
        # See (https://www.mathewmarcus.com/blog/asynchronous-aws-api-requests-with-asyncio.html)
        self.create_clients()

    def create_clients(self):
        lambda_client_future = self.pool.submit(self._create_lambda_client)
        s3_client_future = self.pool.submit(self._create_s3_client)
        dynamodb_client_future = self.pool.submit(self._create_dynamodb_client)
        api_gateway_client_future = self.pool.submit(self._create_api_gateway_client)
        events_client_future = self.pool.submit(self._create_events_client)
        iam_client_future = self.pool.submit(self._create_iam_client)
        iam_resource_future = self.pool.submit(self._create_iam_resource)
        while (not lambda_client_future.done() or not s3_client_future.done() or not dynamodb_client_future.done()
               or not api_gateway_client_future.done() or not events_client_future.done() or not iam_client_future.done() or not iam_resource_future.done()):
            time.sleep(0.1)

    def _create_boto_session(self):
        self.boto_session = boto3.Session()

    def _create_lambda_client(self):
        self.lambda_client = boto3.client("lambda")

    def _create_s3_client(self):
        self.s3_client = boto3.client("s3")

    def _create_dynamodb_client(self):
        self.dynamodb_client = boto3.client("dynamodb")

    def _create_api_gateway_client(self):
        self.api_gateway_client = boto3.client("apigatewayv2")

    def _create_events_client(self):
        self.events_client = boto3.client("events")

    def _create_iam_client(self):
        self.iam_client = boto3.client("iam")
        self.iam = self.iam_client
        # todo: check if the iam from the boto_session is equivalent to the classical iam_client

    def _create_iam_resource(self):
        self.iam_resource = boto3.resource("iam")


if __name__ == "__main__":
    import time
    start_time = time.time()
    clients_instance = CoreClients()
    print(f"Initialization delay = {time.time() - start_time}")
    # print(clients_instance.iam_client.get_role(RoleName="InoftVocalFrameworkLambdaExecution")["Role"])



import json
import os
import time
from concurrent.futures.thread import ThreadPoolExecutor
from pathlib import Path
import logging
import boto3
import botocore
import click
from botocore.exceptions import ClientError, NoCredentialsError

from inoft_vocal_framework.cli.cli_cache import CliCache
from inoft_vocal_framework.skill_builder.skill_settings import Settings

from inoft_vocal_framework.exceptions import raise_if_variable_not_expected_type

from inoft_vocal_framework.cli.core.core_clients import CoreClients
from inoft_vocal_framework.safe_dict import SafeDict


class Core(CoreClients):
    def __init__(self):  # , pool: ThreadPoolExecutor):
        super().__init__()
        self.settings = Settings()
        self.tags = {"framework": "InoftVocalFramework"}

        self._iam_role = None
        self._credentials_arn = None
        self._aws_account_id = None

        self._session_unix_time = None
        self._session_count_created_statement_ids = 0

        self.role_name = "InoftVocalFrameworkLambdaExecution"
        self.http_methods = ['ANY']

        self.default_description = "Created with the Inoft Vocal Framework"

    def get_lambda_function_arn(self, function_name: str):
        return SafeDict(self.lambda_client.get_function(FunctionName=function_name)).get("Configuration").get("FunctionArn").to_str()

    def api_gateway_v2_url(self, api_id: str):
        try:
            return SafeDict(self.api_gateway_client.get_api(ApiId=api_id)).get("ApiEndpoint").to_str(default=None)
        except Exception as e:
            return False

    def get_lambda_function_versions(self, function_name: str) -> list:
        try:
            response = self.lambda_client.list_versions_by_function(FunctionName=function_name)
            return response.get('Versions', list())
        except Exception as e:
            click.echo(f"Lambda function {function_name} not found. Error : {e}")
            return list()

    def create_lambda_function(self, bucket=None, s3_key: str = None, local_zip: bytes = None, function_name: str = None, handler=None,
                               description: str = "Inoft Vocal Framework Deployment",
                               timeout: int = 30, memory_size: int = 512, publish: bool = True, runtime=None):
        """
        Given a bucket and key (or a local path) of a valid Lambda-zip, a function name and a handler, register that Lambda function.
        """
        kwargs = dict(
            FunctionName=function_name,
            Runtime=runtime,
            Role=self.credentials_arn,  # "arn:aws:iam::631258222318:role/service-role/alexa_hackaton-cite-des-sciences_fakenews-challeng-role-ck3sxsyt",  # self.credentials_arn,
            # todo: make dynamic and create automaticly role if missing
            Handler=handler,
            Description=description,
            Timeout=timeout,
            MemorySize=memory_size,
            Publish=publish,
            Layers=[
                "arn:aws:lambda:eu-west-3:631258222318:layer:inoft-vocal-framework_0-38-5:1",
            ]
        )
        if local_zip is not None:
            kwargs['Code'] = {
                'ZipFile': local_zip
            }
        else:
            kwargs['Code'] = {
                'S3Bucket': bucket,
                'S3Key': s3_key
            }

        response = self.lambda_client.create_function(**kwargs)

        lambda_arn = response["FunctionArn"]
        version = response['Version']

        click.echo(f"Created new lambda function {function_name} with arn of {lambda_arn}")
        return lambda_arn

    def update_lambda_function_code(self, lambda_arn: str, object_key_name: str, bucket_name: str):
        response = self.lambda_client.update_function_code(
            FunctionName=lambda_arn,
            S3Bucket=bucket_name,
            S3Key=object_key_name,
            Publish=True,
        )

    def create_api_gateway(self, lambda_arn: str, lambda_name: str, description: str = None) -> str:
        api_name = lambda_name or lambda_arn.split(":")[-1]

        api_creation_response = SafeDict(self.api_gateway_client.create_api(Name=api_name,
            Description=description or self.default_description, Target=lambda_arn, ProtocolType="HTTP"))
        api_id = api_creation_response.get("ApiId").to_str()

        self.create_and_setup_api_routes(api_id=api_id, lambda_arn=lambda_arn)
        return api_id

    def create_and_setup_api_routes(self, api_id: str, lambda_arn: str):
        integration_creation_response = SafeDict(self.api_gateway_client.create_integration(ApiId=api_id,
            IntegrationType="AWS_PROXY", PayloadFormatVersion="2.0", IntegrationUri=lambda_arn, TimeoutInMillis=10000))

        integration_id = integration_creation_response.get("IntegrationId").to_str(default=None)
        if integration_id is None:
            raise Exception(f"Error while creating the route integration of the api towards the lambda_function."
                            f"The integration_id was not found in the integration creation response."
                            f"Please delete the api and redo the deployment."
                            f"If the issue persist, create an Issue topic on the github page of the framework.")

        route_names = ["amazonAlexaV1", "googleAssistantDialogflowV1", "samsungBixbyV1", "appleSiriV1"]
        route_names_to_settings_keys = {"amazonAlexaV1": "alexaApiEndpointUrlNotRecommendedToUse",
                                        "googleAssistantDialogflowV1": "googleAssistantApiEndointUrl",
                                        "samsungBixbyV1": "samsungBixbyApiEndointUrl",
                                        "appleSiriV1": "siriApiEndointUrl"}

        api_gateway_root_url = self.api_gateway_v2_url(api_id=api_id)
        for route_name in route_names:
            self.boto_session.get_credentials()
            response = self.api_gateway_client.create_route(
                ApiId=api_id,
                AuthorizationType="None",
                RouteKey=f"ANY /{route_name}",
                # A space is required between the HTTP method and the /
                Target=f"integrations/{integration_id}",
            )
            self.add_lambda_permission_to_call_api_resource(lambda_arn=lambda_arn, api_id=api_id, route_key=route_name)
            api_route_url = f"{api_gateway_root_url}/{route_name}"
            click.echo(f"Api route {click.style(route_name, fg='green')} creation complete accessible on {api_route_url}")
            self.settings.settings.get_set("deployment", {}).get_set("endpoints", {}).put(
                route_names_to_settings_keys[route_name], api_route_url).reset_navigated_dict()
            # We reset the navigated dict after a final put

    def add_lambda_permission_to_call_api_resource(self, lambda_arn: str, api_id: str, route_key: str):
        response = self.lambda_client.add_permission(
            FunctionName=lambda_arn,
            StatementId=self.get_session_new_statement_id(),
            Action="lambda:InvokeFunction",
            Principal="apigateway.amazonaws.com",
            SourceArn=f"arn:aws:execute-api:{self.boto_session.region_name}:{self.aws_account_id}:{api_id}/*/*/{route_key}",
        )

    @property
    def iam_role(self) -> SafeDict:
        if self._iam_role is None:
            try:
                self._iam_role = SafeDict(self.iam.get_role(RoleName=self.role_name)["Role"])
            except botocore.exceptions.ClientError:
                # If a ClientError happened while getting the role, it means he do not exist.
                logging.debug(f"Creating {self.role_name} IAM Role..")

                attach_policy_obj = json.loads(self.attach_policy)
                assume_policy_obj = json.loads(self.assume_policy)

                self._iam_role = SafeDict(self.iam.create_role(RoleName=self.role_name, AssumeRolePolicyDocument=self.assume_policy)["Role"])

                # create or update the role's policies if needed
                policy = self.iam_resource.RolePolicy(self.role_name, "inoft-vocal-permissions")
                try:
                    if policy.policy_document != attach_policy_obj:
                        logging.debug(f"Updating inoft-vocal-permissions policy on {self.role_name} IAM Role.")
                        policy.put(PolicyDocument=self.attach_policy)

                except botocore.exceptions.ClientError:
                    logging.debug(f"Creating inoft-vocal-permissions policy on {self.role_name} IAM Role.")
                    policy.put(PolicyDocument=self.attach_policy)

                role_policy_dict = self._iam_role.get("AssumeRolePolicyDocument").to_dict()
                if role_policy_dict != assume_policy_obj:
                    if (SafeDict(role_policy_dict["Statement"][0]).get("Principal").get("Service").to_any()
                            != assume_policy_obj["Statement"][0]["Principal"]["Service"]):

                        logging.debug(f"Updating assume role policy on {self.role_name} IAM Role.")
                        self.iam_client.update_assume_role_policy(RoleName=self.role_name, PolicyDocument=self.assume_policy)

        return self._iam_role

    @property
    def credentials_arn(self):
        if self._credentials_arn is None:
            self._credentials_arn = self.iam_role.get("Arn").to_str()
        return self._credentials_arn

    @credentials_arn.setter
    def credentials_arn(self, credentials_arn) -> None:
        self._credentials_arn = credentials_arn

    @property
    def aws_account_id(self) -> str:
        if self._aws_account_id is None:
            self._aws_account_id = boto3.client("sts").get_caller_identity().get("Account")
        return self._aws_account_id

    def get_session_new_statement_id(self) -> str:
        # No restriction is set on the statement id. To created my type of ids, i take the same unix
        # time across the session, and each time we generated a new statement id, i add 1 to the unix time.
        # Dumb and simple, but it assure that i will never have the same id twice.
        if self._session_unix_time is None:
            self._session_unix_time = time.time()
        self._session_count_created_statement_ids += 1

        # After turning the modified unix time into a string, we remove any point, that would have
        # been used to  separate the decimals, because AWS cannot accept a point in a id string.
        return str(self._session_unix_time + self._session_count_created_statement_ids).replace(".", "")

    def create_s3_bucket_if_missing(self, bucket_name: str, region_name: str):
        try:
            self.s3_client.head_bucket(Bucket=bucket_name)
        except ClientError as e:
            click.echo(f"Trying to create a new S3 Bucket with name {click.style(text=bucket_name, fg='yellow', bold=True)}"
                  f" in region {click.style(text=region_name, fg='yellow', bold=True)}")
            available_regions_for_s3 = self.boto_session.get_available_regions(service_name="s3")
            if region_name not in available_regions_for_s3:
                raise Exception(f"The region {region_name} was not available for s3. Here is the available regions : {available_regions_for_s3}")

            self.s3_client.create_bucket(Bucket=bucket_name, CreateBucketConfiguration={"LocationConstraint": region_name})
            click.echo(f"Completed creation of the new bucket.")

    def upload_to_s3(self, filepath: str, object_key_name: str, bucket_name: str, region_name: str) -> bool:
        # If an error happen while uploading to S3, then the upload will not be
        # successful. We use that as our way to send a success response.
        try:
            self.create_s3_bucket_if_missing(bucket_name=bucket_name, region_name=region_name)
            self.s3_client.upload_file(Filename=filepath, Bucket=bucket_name, Key=object_key_name)
            return True
        except NoCredentialsError as e:
            click.echo(f"Almost there ! You just need to configure your AWS credentials."
                  f"\nYou can follow the official documentation (you will need to install the awscli by running pip install awscli) "
                  f"then go to https://boto3.amazonaws.com/v1/documentation/api/latest/guide/quickstart.html#configuration"
                  f"\nOr you can follow a video made by Robinson Labourdette from Inoft, as a guide to configure your credentials.")
            if click.confirm("Press Y to access the video."):
                selected_language = click.prompt("Type the language in which you would like the video. The followings are available :", type=click.Choice(["English", "French"]))
                if selected_language == "English":
                    click.echo("English here you go !")
                elif selected_language == "French":
                    click.echo("Français la voila !")
                click.echo("Follow the instructions that is available in the video, then when you are all set up, redo the command you just tried")
                while True:
                    if click.confirm("Press y to exit"):
                        exit(201)
                        break
                        # A little break, just to make sure that if the exit do not work, we do not stay stuck in this loop.
            else:
                click.echo("Ok, then follow the boto documentation, and once you have configured your credentials, relaunch the cli command that you were trying to use"
                      " (https://boto3.amazonaws.com/v1/documentation/api/latest/guide/quickstart.html#configuration)")
        except Exception as e:
            click.echo(f"Error while getting/creating/uploading to the S3 bucket : {e}")
            return False

    def remove_from_s3(self, file_name: str, bucket_name: str):
        try:
            self.s3_client.head_bucket(Bucket=bucket_name)
        except botocore.exceptions.ClientError as e:
            # If a client error is thrown, then check that it was a 404 error.
            # If it was a 404 error, then the bucket does not exist.
            error_code = int(e.response["Error"]["Code"])
            if error_code == 404:
                return False

        try:
            self.s3_client.delete_object(Bucket=bucket_name, Key=file_name)
            return True
        except (botocore.exceptions.ParamValidationError, botocore.exceptions.ClientError):
            return False

if __name__ == "__main__":
    Core().upload_to_s3("F:\Inoft\hackaton cite des sciences 1\lambda_project\inoft_vocal_framework\cli\core\core_clients.py",
                        "letestduframeworkinoft", "eu-west-3")



import inspect
import os
import time
import zipfile
from concurrent.futures.thread import ThreadPoolExecutor
from pathlib import Path

import boto3
import botocore
import click
from botocore.exceptions import ClientError
from click import ClickException

import inoft_vocal_framework
from inoft_vocal_framework.cli.aws_utils import raise_if_bucket_name_not_valid
from inoft_vocal_framework.cli.cli_cache import CliCache
from inoft_vocal_framework.cli.core.core import Core
from inoft_vocal_framework.skill_builder.skill_settings import Settings


class DeployHandler(Core):
    def __init__(self):
        click.echo("Initializing AWS clients... (this can take a few seconds)")
        start_time = time.time()
        super().__init__()
        click.echo(f"Took {click.style(text=f'{round(time.time() - start_time, 2)}s', fg='white')} to initiate AWS clients")
        # todo: make the initialization even more asynchronous, by calling the handle function while the AWS ressources are initializing.
        #  Right now, the initialization of the resources are async, but we must wait for their completion in order to call the handle function.

    """def handle(self):
        # result = self.pool.submit(super().__init__, {"pool": self.pool})
        self._handle()"""

    def handle(self):
        changed_root_folderpath = False
        app_project_root_folderpath = CliCache.cache().get("last_app_project_root_folderpath").to_str(default=None)

        def prompt_user_to_select_folderpath():
            return click.prompt(text="What is the root folder path of your project ? "
                                     "This is the default if you do not write anything :",
                                default=str(Path(os.path.dirname(os.path.realpath(inoft_vocal_framework.__file__))).parent))

        if app_project_root_folderpath is None:
            app_project_root_folderpath = prompt_user_to_select_folderpath()
            changed_root_folderpath = True
        else:
            if not click.confirm(f"Do you want to deploy the project that is present in the following folder : {app_project_root_folderpath}"):
                app_project_root_folderpath = prompt_user_to_select_folderpath()
                changed_root_folderpath = True

        while not os.path.exists(app_project_root_folderpath):
            if click.confirm(text="The root folder path of the project has not been found."
                                  "Do you want to select a new folderpath ? Otherwise the CLI will close."):
                app_project_root_folderpath = prompt_user_to_select_folderpath()
                changed_root_folderpath = True
            else:
                exit(200)

        if changed_root_folderpath is True:
            CliCache.cache().put("last_app_project_root_folderpath", app_project_root_folderpath)
            CliCache.save_cache_to_yaml()
            click.echo(f"Saved the folderpath of your project for {click.style(text='faster load next time ;)', fg='blue')}")

        self.settings.find_load_settings_file(root_folderpath=app_project_root_folderpath)

        # I cannot create a variable that contain the deployment settings, otherwise it will be hell with the resets of the safedict.
        bucket_name = self.settings.settings.get_set("deployment", {}).get("s3_bucket_name").to_str(default=None)
        if bucket_name is None:
            bucket_name = click.prompt("What will be your S3 bucket name to host your deployment files ?", type=str)
            self.settings.settings.get_set("deployment", {}).put("s3_bucket_name", bucket_name).reset_navigated_dict()
        raise_if_bucket_name_not_valid(bucket_name=bucket_name)

        lambda_name = self.settings.settings.get_set("deployment", {}).get("lambda_name").to_str(default=None)
        if lambda_name is None:
            lambda_name = click.prompt("What will be your Lambda (which act as the 'server' for your app) name ?", type=str)
            self.settings.settings.get_set("deployment", {}).put("lambda_name", lambda_name).reset_navigated_dict()

        self.settings.save_settings()
        click.echo(f"Saved your settings to your app_settings file at : {click.style(text=str(self.settings.last_settings_filepath), fg='yellow')}")

        # todo: ask and check lambda handler file/function path

        self.deploy( app_project_root_folderpath=app_project_root_folderpath, bucket_name=bucket_name,
                     lambda_name=lambda_name, lambda_handler="app.lambda_handler", upload_zip=True)

    def deploy(self, app_project_root_folderpath: str, bucket_name: str, lambda_name: str, lambda_handler: str,
               upload_zip: bool = True, app_project_existing_zip_filepath: str = None,
               lambda_description: str = "Inoft Vocal Framework Deployment",
               lambda_timeout_seconds=30, lambda_memory_size=512, publish=True, runtime="python3.7"):

        if self.settings.settings_loaded is False:
            self.settings.find_load_settings_file(root_folderpath=app_project_root_folderpath)

        # Make sure this isn't already deployed.
        """deployed_versions = self.get_lambda_function_versions(lambda_name)
        if len(deployed_versions) > 0:
            click.echo(f"This lambda function for this application is {click.style('already deployed', fg='green')}" +
                  f" If some part of the full deployment failed or you changed some resource name, and you want to do it again, type Y."
                  f" Otherwise, use the {click.style('inoft update', bold=True)} command")
            if not click.confirm("Do you want to redo the deployment ?"):
                click.echo(f"Ok ! Remember to use the {click.style('inoft update', bold=True)} command !")
                exit()
        """

        zip_filepath, upload_success = None, None
        if upload_zip is True:
            if app_project_existing_zip_filepath is None:
                # Create the Lambda Zip
                zip_filepath = self.create_package(app_folder_path=app_project_root_folderpath)
            else:
                if not os.path.isfile(app_project_existing_zip_filepath):
                    raise Exception(f"Existing app project zip file do not exist at filepath : {app_project_existing_zip_filepath}")
                zip_filepath = app_project_existing_zip_filepath

            # Upload it to S3
            click.echo("Uploading the app zip file to S3")
            upload_success = self.upload_to_s3(filepath=zip_filepath, object_key_name=Path(zip_filepath).name,
                                               bucket_name=bucket_name, region_name="eu-west-3")
            if not upload_success:
                raise ClickException("Unable to upload to S3. Look in the logs what caused the errors,"
                                     " and modify your app_settings file in order to fix the issue."
                                     "\nCommon issues include :"
                                     "\n - Not having setup your credentials correctly (run the command : aws configure)"
                                     "\n - Trying to use a S3 bucket name that is not available."
                                     "\n - Having modified the app_settings file, but not have saved it with CTRL+S")
            click.echo("Uploading completed.")

        try:
            lambda_arn = self.get_lambda_function_arn(function_name=lambda_name)
            click.echo(f"Using the existing lambda function {lambda_name}")
            if upload_success is True:
                self.update_lambda_function_code(lambda_arn=lambda_arn, object_key_name=Path(zip_filepath).name, bucket_name=bucket_name)
                click.echo(f"Updated the code of the lambda with arn {lambda_arn}")

        except botocore.exceptions.ClientError:
            try:
                lambda_arn = self.create_lambda_function(bucket=bucket_name, s3_key=Path(zip_filepath).name, function_name=lambda_name,
                                                         handler=lambda_handler, description=lambda_description,
                                                         timeout=lambda_timeout_seconds, memory_size=lambda_memory_size,
                                                         runtime=runtime, publish=publish)
                click.echo(f"Created the lambda function : {lambda_name}")
            except Exception as e:
                click.echo(f"Error while updating the function by giving it a S3 path. Trying to update the"
                      f"function with a byte_stream of the zip file of the deployment package : {e}")

                with open(zip_filepath, mode="rb") as file_stream:
                    byte_stream = file_stream.read()

                lambda_arn = self.create_lambda_function(local_zip=byte_stream, function_name=lambda_name,
                                                         handler=lambda_handler, description=lambda_description,
                                                         timeout=lambda_timeout_seconds, memory_size=lambda_memory_size,
                                                         runtime=runtime, publish=publish)
                click.echo(f"Created the lambda function : {lambda_name}")

        # Create and configure the API Gateway
        need_to_create_api = False
        api_gateway_id = self.settings.settings.get("deployment").get("apiGatewayId").to_str(default=None)
        if api_gateway_id is None or self.api_gateway_v2_url(api_gateway_id) is None:
            api_id = self.create_api_gateway(lambda_arn=lambda_arn, lambda_name=lambda_name)
            click.echo(f"Created a new {click.style(text='API Gateway', bold=True)}")
            self.settings.settings.get("deployment").put("apiGatewayId", api_id)
            self.settings.save_settings()
        else:
            click.echo(f"Using the existing ApiGatewayV2 with id {click.style(text=api_gateway_id, bold=True, fg='green')}")

        return True

        # Remove the project zip from S3.
        if not source_zip:
            self.remove_uploaded_zip()


    @staticmethod
    def create_package(app_folder_path: str) -> str:
        # todo: fix issue where we have the following printed for each single file (most likely come from the fact that i do not delete the old zip before creating a new one)
        #  return self._open_to_write(zinfo, force_zip64=force_zip64)
        #  c:\users\labourdette\anaconda3\lib\zipfile.py:1473: UserWarning: Duplicate name: 'inoft_vocal_framework/utils/__init__.py'

        archive_destination_filepath = os.path.join(Path(app_folder_path).parent, f"{Path(app_folder_path).name}.zip")
        click.echo(f"Making an archive from all the files and folders in {app_folder_path} to {archive_destination_filepath}")

        folders_names_to_excludes = [".aws-sam", ".idea", "__pycache__", "venv"]

        with zipfile.ZipFile(archive_destination_filepath, "w") as zip_object:
            has_found_framework_in_project_files = False

            # Include the projects files
            for root_dirpath, dirs, filenames in os.walk(app_folder_path, topdown=True):
                # The topdown arg allow use to modify the dirs list in the walk, and so we can easily exclude folders.
                dirs[:] = [dirpath for dirpath in dirs if Path(dirpath).name not in folders_names_to_excludes]

                if Path(root_dirpath).name == "inoft_vocal_framework":
                    has_found_framework_in_project_files = True

                relative_root_dirpath = root_dirpath.replace(app_folder_path, "")
                for filename in filenames:
                    zip_object.write(filename=os.path.join(root_dirpath, filename),
                                     arcname=os.path.join(relative_root_dirpath, filename))

            # todo: when doing a redeploy, check that the lambda layer used, is the right layer for the current framework version

            if has_found_framework_in_project_files is False:
                # On the condition that we have not found the framework in the project files. This function exist both for development sake,
                # where i can deploy with a dev version of the framework without having to publish to pip, while stile using the command line
                # interface from the pip package. And also since the framework (and its correct version) will be included in the deployment
                # package, so that if an user download an app deployed in a different version that the current version of the framework he
                # is using, the version of the framework that will be used will be the one included in its package, not the one installed.

                # We will include the inoft_vocal_framework himself (he is not included in the lambda layers)
                # I do not include the framework in the lambda layer, because it would be weird for someone to update the framework,
                # do a redeploy, and not have upgraded its deploy. Where as we can check if its layer is the right layer for the version
                # of his framework. And mostly because it would be annoying for me to recreate a new lambda layer on each update of the
                # framework, and that he is light enough to be included in the package and the increase in upload time to not be noticeable ;)
                inoft_vocal_framework_folder_path = os.path.dirname(os.path.realpath(inoft_vocal_framework.__file__))
                for root_dirpath, dirs, filenames in os.walk(inoft_vocal_framework_folder_path):
                    relative_dirpath_with_root_included = root_dirpath.replace(str(Path(inoft_vocal_framework_folder_path).parent), "")
                    # We only replace the parent of the framework folder path, because we want it to be in its inoft_vocal_framework folder.
                    for filename in filenames:
                        zip_object.write(filename=os.path.join(root_dirpath, filename),
                                         arcname=os.path.join(relative_dirpath_with_root_included, filename))


        # from shutil import make_archive
        # zip_filepath = make_archive(base_name=archive_destination_filepath_without_extension, format="zip", root_dir=app_folder_path)
        click.echo(f"Archive completed at {click.style(text=archive_destination_filepath, bold=True)}")

        # Warn if this is too large for Lambda.
        file_stats = os.stat(archive_destination_filepath)
        if file_stats.st_size > 52428800:
            click.echo("Warning: Application zip package is likely to be too large for AWS Lambda. Try to make it smaller")

        return archive_destination_filepath


if __name__ == "__main__":
    DeployHandler().handle()
